import sys
import time
import argparse
from datetime import datetime, date, timedelta

from kabaret.app.session import KabaretSession
from kabaret.app.actors.flow import Flow

from libreflow.utils.kabaret.jobs.jobs_actor import Jobs


class CleanJobsSession(KabaretSession):

    @staticmethod
    def parse_command_line_args(args):
        base_args = KabaretSession.parse_command_line_args(args)
        base_args, remaining_args = base_args[:-1], base_args[-1]

        parser = argparse.ArgumentParser(
            description='Jobs Cleaning Session Arguments'
        )
        parser.add_argument(
            '-c', '--clean-time', type=str, default='13:00:00', dest='clean_time'
        )
        parser.add_argument(
            '-n', '--nb-days-before', type=int, default=1, dest='nb_days_before'
        )
        parser.add_argument(
            '-l', '--limit-time', type=str, default='19:00:00', dest='limit_time'
        )
        parser.add_argument(
            '-e', '--exclude-day', action='append', type=int, dest='exclude_days'
        )
        parser.add_argument(
            '-v', '--verbose', action='store_true', dest='verbose'
        )

        values, remaining_args = parser.parse_known_args(remaining_args)
        
        return base_args + (
            values.clean_time,
            values.nb_days_before, 
            values.limit_time,
            values.exclude_days,
            values.verbose,
            remaining_args
         )
    
    def __init__(self, session_name=None, debug=False, clean_time='13:00:00', nb_days_before=1, limit_time='19:00:00', exclude_days=None, verbose=False):
        super(CleanJobsSession, self).__init__(session_name, debug)
        self._clean_time = clean_time
        self._nb_days_before = nb_days_before
        self._limit_time = limit_time
        self._verbose = verbose

        if exclude_days is None:
            self._exclude_days = []
        else:
            self._exclude_days = exclude_days
    
    def _create_actors(self):
        Flow(self)
        Jobs(self)
    
    def start(self):
        if self._verbose:
            self.log_info('CLEANED JOBS')
            self.log_info(' %s | %s | %s | %s' % (
                'JOB ID'.center(36),
                'EMITTED ON'.center(17),
                'BY'.center(20),
                'DESCRIPTION'.center(97)
            ))
            self.log_info('-' * 180)
        
        try:
            if time.time() > self.get_clean_time_of_current_day():
                self.clean_jobs(self.get_limit_time())

            while True:
                time.sleep(self.get_next_clean_time() - time.time())

                if self.current_day_is_valid():
                    self.clean_jobs(self.get_limit_time())
        except KeyboardInterrupt:
            self.log_info('Cleaner stopped. Exiting...')
            return
    
    def get_limit_time(self):
        today = date.today()
        limit_day = (today - timedelta(days=self._nb_days_before)).strftime('%y-%m-%d')
        limit_timestamp = time.mktime(time.strptime('%s %s' % (limit_day, self._limit_time), '%y-%m-%d %H:%M:%S'))

        return limit_timestamp
    
    def get_clean_time_of_current_day(self):
        day = date.today()
        clean_timestamp = time.mktime(time.strptime('%s %s' % (day.strftime('%y-%m-%d'), self._clean_time), '%y-%m-%d %H:%M:%S'))
        
        return clean_timestamp

    def get_next_clean_time(self):
        clean_timestamp = self.get_clean_time_of_current_day()

        if time.time() > clean_timestamp:
            clean_timestamp += 86400.0
        
        return clean_timestamp
    
    def current_day_is_valid(self):
        return date.today().weekday() not in self._exclude_days
    
    def clean_jobs(self, timestamp):
        if self._verbose:
            last = datetime.fromtimestamp(time.time()).strftime('%y-%m-%d %H:%M:%S')
            next = datetime.fromtimestamp(self.get_next_clean_time()).strftime('%y-%m-%d %H:%M:%S')
            self.log_info(('Last clean: %s' % last).rjust(180))
            self.log_info(('Next clean: %s' % next).rjust(180))
        
        for j in self.cmds.Jobs.list_jobs():
            if j['created_on'] < timestamp:
                if self._verbose:
                    self.log_info(' %s | %s | %s | %s' % (
                        j['jid'],
                        datetime.fromtimestamp(j['created_on']).strftime('%y-%m-%d %H:%M:%S'),
                        j['creator'].center(20),
                        j['job_label']
                    ))
                
                self.cmds.Jobs.delete_job(j['jid'])


def main(argv):
    (
        session_name,
        host,
        port,
        cluster_name,
        db,
        password,
        debug,
        clean_time,
        nb_days_before,
        limit_time,
        exclude_days,
        verbose,
        _,
    ) = CleanJobsSession.parse_command_line_args(argv)

    session = CleanJobsSession(session_name=session_name, debug=debug, clean_time=clean_time, nb_days_before=nb_days_before, limit_time=limit_time, exclude_days=exclude_days, verbose=verbose)
    session.cmds.Cluster.connect(host, port, cluster_name, db, password)

    session.start()


if __name__ == "__main__":
    main(sys.argv[1:])


import sys
from string import Formatter
from collections import OrderedDict

from kabaret.flow_contextual_dict import get_contextual_dict


def keywords_from_format(format):
    return [fname for _, fname, _, _ in Formatter().parse(format) if fname]


def _get_param_values(leaf, name, level=0, max_level=sys.maxsize):
    session = leaf.root().session()
    param_oid = leaf.oid() + '/' + name
    values = []
    
    if session.cmds.Flow.exists(param_oid):
        value = session.cmds.Flow.get_value(param_oid)
        values.append(value)
    
    parent = leaf._mng.parent
    
    if parent is not leaf.root() and level < max_level:
        parent_values = _get_param_values(parent, name, level + 1, max_level)
        return parent_values + values
    else:
        return values


def get_context_value(
        leaf,
        param_name,
        delim='',
        context_name='settings',
        max_level=sys.maxsize,
    ):
    
    session = leaf.root().session()
    
    # Recursively search param
    param_values = _get_param_values(leaf, param_name, max_level=max_level)
    
    if not param_values:
        return None
    
    # Build context value from accumulated values
    context_value = delim.join(param_values)
    
    # Search for contextual keys
    try:
        keys = keywords_from_format(context_value)
    except ValueError:
        session.log_error((
            f"Invalid value format \"{context_value}\"\n\n"
            "<<<<<<<<<<< TRACEBACK >>>>>>>>>>\n"
        ))
        raise
    
    if keys:
        # Get contextual values
        context = get_contextual_dict(leaf, context_name)
        values = OrderedDict()
        
        for key in keys:
            try:
                values[key] = context[key]
            except KeyError:
                session.log_error((
                    f"No key \"{key}\" in {leaf.oid()} context\n\n"
                    "<<<<<<<<<<< TRACEBACK >>>>>>>>>>\n"
                ))
                raise
        
        context_value = context_value.format(**values)
    
    return context_value



import os
import gazu
import re
from fnmatch import fnmatch

from kabaret import flow
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict

from .departments import Department
from .maputils import ItemMap, CreateItemAction, ClearMapAction
from .lib import AssetDependency, DropAssetAction  # , KitsuSettingsView
from .kitsu import KitsuSequence, KitsuShot, UpdateItemsKitsuSettings
from .site import RequestRevisions, MultiRequestRevisions, MultiRequestedRevisions, SiteJobsPoolNames, ActiveSiteChoiceValue, ActiveSitesMultichoiceValue, RequestedRevisions
from .dependency import get_dependencies
from .file import GenericRunAction
from .users import ToggleBookmarkAction


class Casting(flow.Map):

    ICON = ("icons.flow", "casting")

    drag_assets = flow.Child(DropAssetAction)

    @classmethod
    def mapped_type(cls):
        return AssetDependency

    def columns(self):
        return ["Name", "Description"]

    def row(self, item):
        _, row = super(Casting, self).row(item)

        return item.get().oid(), row

    def _fill_row_cells(self, row, item):
        asset = item.get()
        row["Name"] = asset.id.get()
        row["Description"] = asset.description.get()


class DisplayKitsuSettings(flow.Action):

    _map = flow.Parent()

    def needs_dialog(self):
        return False

    def allow_context(self, context):
        return context and context.endswith(".inline")

    def run(self, button):
        displayed = self._map._display_kitsu_settings.get()
        self._map._display_kitsu_settings.set(not displayed)
        self._map.touch()


class ShotDepartments(flow.Object):

    layout = flow.Child(Department).ui(expanded=True)
    animation = flow.Child(Department).ui(expanded=True)
    compositing = flow.Child(Department).ui(expanded=True)


class ShotElements(flow.values.MultiChoiceValue):
    
    def choices(self):
        template = self.root().project().admin.dependency_templates['shot']
        choices = []
        
        for element, data in template.get_dependencies().items():
            if data.get('requestable', True):
                choices.append(element)
            
        return choices


class DependencyOptions(flow.Object):
    
    request_dependencies = flow.SessionParam(True).ui(editor='bool')
    include_last_revisions = flow.SessionParam(False).ui(
        editor='bool',
        tooltip='Include the latest revision of each dependency')


class ShotElementSelection(flow.Object):
    
    _shot = flow.Parent(2)
    _sequence = flow.Parent(4)
    select_all_elements = flow.SessionParam(False).ui(editor='bool').watched()
    elements = flow.Param([], ShotElements).watched()
    
    def __init__(self, parent, name):
        super(ShotElementSelection, self).__init__(parent, name)
        self._kitsu_casting = None
    
    def _get_casting(self):
        bindings = self.root().project().kitsu_bindings()
        casting = bindings.get_shot_casting(
            self._shot.name(),
            self._sequence.name()
        )

        return casting
    
    def ensure_kitsu_casting(self):
        if self._kitsu_casting is None:
            self._kitsu_casting = self._get_casting()
        
        return self._kitsu_casting
    
    def child_value_changed(self, child_value):
        if child_value is self.select_all_elements:
            if self.select_all_elements.get():
                self.elements.set(self.elements.choices())
            else:
                self.elements.set([])
    
    def compute_oids(self, include_dependencies=False, latest_dependencies=False):
        elements = self.elements.get()
        template = self.root().project().admin.dependency_templates['shot']
        bindings = self.root().project().kitsu_bindings()
        deps = template.get_dependencies()
        
        kitsu_casting = self.ensure_kitsu_casting()
        oids = set()
        
        for element in self.elements.get():
            dep_data = deps[element]
            kitsu_data = dep_data.get('kitsu', None)
            
            if kitsu_data is not None and kitsu_data.get('entity', None) == 'Asset':
                # Assets
                for asset_name, asset_data in kitsu_casting.items():
                    asset_type = asset_data['type']
                    
                    if asset_type != kitsu_data['type']:
                        continue
                    
                    files_data = template.get_dependency_files(element)
                    asset_oid = bindings.get_asset_oid(asset_name)
                    
                    # Get files
                    for file_name, file_data in files_data.items():
                        rev_oids = self._get_revision_oids(asset_oid, file_name, file_data, include_dependencies, latest_dependencies)
                        # pprint(oids)
                        oids = oids.union(set(rev_oids))
            else:
                # Shots
                files_data = template.get_dependency_files(element)
                
                for file_name, file_data in files_data.items():
                    rev_oids = self._get_revision_oids(self._shot.oid(), file_name, file_data, include_dependencies, latest_dependencies)
                    oids = oids.union(set(rev_oids))
        
        return sorted(list(oids))
    
    def _get_revision_oids(self, root_oid, file_name, file_data, include_dependencies=False, latest_dependencies=False):
        # Check if given file exists
        file_oid = "%s/departments/%s/files/%s" % (
            root_oid,
            file_data['department'],
            file_name.replace('.', '_')
        )
        
        if not self.root().session().cmds.Flow.exists(file_oid):
            self.root().session().log_warning('Unknown file with oid ' + file_oid)
            return []
        
        revision_name = file_data.get('revision', '[last]')
        revision_oids = [file_oid + '/history/revisions/' + revision_name]
        
        # Return if dependencies are not requested
        if not include_dependencies:
            return revision_oids
        
        if revision_name == '[last]':
            revision_name = None
        
        # Get dependencies
        
        file_object = self.root().get_object(file_oid)
        dependencies = get_dependencies(
            leaf=file_object,
            predictive=True,
            real=True,
            revision_name=revision_name)
        
        for d in dependencies:
            if not d['in_breakdown']:
                self.root().session().log_warning(
                    'Dependency %s not in breakdown ' % d['entity_oid']
                )
                continue
            
            revision_oid = d['revision_oid']
            
            if revision_oid is None:
                self.root().session().log_warning((
                    'Dependency %s in breakdown '
                    'but not defined in the flow' % d['entity_oid']
                ))
                continue
            
            revision_oids.append(revision_oid)
            
        # Get predictive dependencies if required
        if latest_dependencies:
            real_deps = get_dependencies(
                leaf=file_object,
                predictive=True,
                revision_name=revision_name)
            
            for d in real_deps:
                revision_oid = d['revision_oid']
                
                if revision_oid is not None and not revision_oid in revision_oids:
                    revision_oids.append(revision_oid)
        
        return revision_oids


class RequestShot(RequestRevisions):
    
    _shot = flow.Parent()
    _sequence = flow.Parent(3)
    selection = flow.Child(ShotElementSelection).ui(expanded=True)
    dependency_options = flow.Child(DependencyOptions).ui(expanded=True)
    revisions = flow.Child(RequestedRevisions).ui(expanded=True)
    pattern = flow.SessionParam('').ui(
        placeholder='Revision oid pattern',
        hidden=True)
    
    def get_buttons(self):
        self.message.set('<h2>Request %s%s</h2>' % (self._sequence.name(), self._shot.name()))
        self.selection.ensure_kitsu_casting()
        
        return super(RequestShot, self).get_buttons()
    
    def run(self, button):
        if button == 'Close':
            return
        elif button == 'Request':
            return super(RequestShot, self).run(button)
        
        oids = self.selection.compute_oids(
            include_dependencies=self.dependency_options.request_dependencies.get(),
            latest_dependencies=self.dependency_options.include_last_revisions.get()
        )
        self.pattern.set(';'.join(oids))
        
        return super(RequestShot, self).run(button)


class MultiRequestShot(MultiRequestRevisions):

    _shot = flow.Parent()
    _sequence = flow.Parent(3)
    selection = flow.Child(ShotElementSelection).ui(expanded=True)
    dependency_options = flow.Child(DependencyOptions).ui(expanded=True)
    revisions = flow.Child(MultiRequestedRevisions).ui(expanded=True)
    pattern = flow.SessionParam('').ui(
        placeholder='Revision oid pattern',
        hidden=True)
    
    def get_buttons(self):
        self.message.set('<h2>Request %s%s towards multiple sites</h2>' % (self._sequence.name(), self._shot.name()))

        return super(MultiRequestShot, self).get_buttons()
    
    def run(self, button):
        if button == 'Close':
            return
        elif button == 'Request':
            return super(MultiRequestShot, self).run(button)
        
        oids = self.selection.compute_oids(
            include_dependencies=self.dependency_options.request_dependencies.get(),
            latest_dependencies=self.dependency_options.include_last_revisions.get()
        )
        self.pattern.set(';'.join(oids))
        
        return super(MultiRequestShot, self).run(button)


class PlayblastChoiceValue(flow.values.ChoiceValue):
    
    CHOICES = ['Layout', 'Blocking', 'Animation', 'Compositing']

    def __init__(self, parent, name):
        super(PlayblastChoiceValue, self).__init__(parent, name)
        self._to_update = True
        self._exclude_choice = None

    def choices(self):
        choices = self.CHOICES.copy()

        if self._exclude_choice is not None:
            choices.remove(self._exclude_choice)
        
        return choices
    
    def exclude_choice(self, choice):
        self._exclude_choice = choice
        self.touch()
    
    def update(self, exclude_choice=None):
        self._exclude_choice = exclude_choice
        self._to_update = False

        if self.get() == exclude_choice:
            self.set(self.choices()[0])
            self._to_update = True
        
        self.touch()
    
    def notify(self):
        if self._watched and self._to_update:
            self._mng.parent.child_value_changed(self)


class ComparePreviews(GenericRunAction):

    ICON = ('icons.libreflow', 'compare-previews')

    _shot = flow.Parent()

    task_1 = flow.Param('Animation', PlayblastChoiceValue).watched()
    task_2 = flow.Param('Compositing', PlayblastChoiceValue).watched()

    FILE_BY_TASK = {
        'Layout': '/departments/layout/files/layout_movie_mov',
        'Blocking': '/departments/animation/files/blocking_movie_mov',
        'Animation': '/departments/animation/files/animation_movie_mov',
        'Compositing': '/departments/compositing/files/compositing_movie_mov'
    }

    def runner_name_and_tags(self):
        return 'RV', []
    
    def get_run_label(self):
        return 'Compare previews'

    def get_buttons(self):
        self.message.set('<h2>Compare previews</h2>')
        self.task_1.revert_to_default()
        self.task_2.revert_to_default()

        return ['Compare', 'Close']
    
    def get_version(self, button):
        return super(GenericRunAction, self).get_version(button)
    
    def needs_dialog(self):
        return True
    
    def extra_argv(self):
        return ['-wipe', '-autoRetime', '0'] + self._previews
    
    def child_value_changed(self, child_value):
        if child_value is self.task_1:
            self.task_2.update(exclude_choice=self.task_1.get())
            self.task_1.exclude_choice(self.task_2.get())
        elif child_value is self.task_2:
            self.task_1.update(exclude_choice=self.task_2.get())
            self.task_2.exclude_choice(self.task_1.get())
    
    def _get_preview_path(self, preview_oid):
        path = None

        if self.root().session().cmds.Flow.exists(preview_oid):
            preview = self.root().get_object(preview_oid)
            head = preview.get_head_revision()

            if head is not None and head.get_sync_status() == 'Available':
                path = head.get_path()
        
        return path
    
    def run(self, button):
        if button == 'Close':
            return
        
        preview_1_oid = self._shot.oid() + self.FILE_BY_TASK[self.task_1.get()]
        preview_2_oid = self._shot.oid() + self.FILE_BY_TASK[self.task_2.get()]

        preview_1_path = self._get_preview_path(preview_1_oid)

        if preview_1_path is None:
            self.message.set((
                '<h2>Compare previews</h2>'
                f'<font color=#D5000D>{self.task_1.get()} preview not available !'
            ))
            return self.get_result(close=False)
        
        preview_2_path = self._get_preview_path(preview_2_oid)

        if preview_2_path is None:
            self.message.set((
                '<h2>Compare previews</h2>'
                f'<font color=#D5000D>{self.task_2.get()} preview not available !'
            ))
            return self.get_result(close=False)
        
        self._previews = [preview_1_path, preview_2_path]
        result = super(ComparePreviews, self).run(button)

        return self.get_result(close=False)


class Shot(KitsuShot):

    ICON = ("icons.flow", "shot")

    _sequence = flow.Parent(2)
    settings = flow.Child(ContextualView).ui(hidden=True)
    # casting = flow.Child(Casting)

    description = flow.Param("")
    departments = flow.Child(ShotDepartments).ui(expanded=True)
    
    request = flow.Child(RequestShot)
    multi_request = flow.Child(MultiRequestShot)
    compare = flow.Child(ComparePreviews)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(shot=self.name())


class Shots(ItemMap):

    ICON = ("icons.flow", "shot")

    item_prefix = "p"

    _display_kitsu_settings = flow.BoolParam(False)

    with flow.group("Kitsu"):
        toggle_kitsu_settings = flow.Child(DisplayKitsuSettings)
        update_kitsu_settings = flow.Child(UpdateItemsKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return flow.injection.injectable(Shot)

    def columns(self):
        names = ["Name"]

        if self._display_kitsu_settings.get():
            names.extend(
                ["Movement", "Nb frames", "Frame in", "Frame out", "Multiplan"]
            )

        return names

    def _fill_row_cells(self, row, item):
        row["Name"] = item.name()

        if self._display_kitsu_settings.get():
            row["Nb frames"] = item.kitsu_settings["nb_frames"].get()

            data = item.kitsu_settings["data"].get()

            row["Movement"] = data["movement"]
            row["Frame in"] = data["frame_in"]
            row["Frame out"] = data["frame_out"]
            row["Multiplan"] = data["multiplan"]


class SequenceElements(flow.values.MultiChoiceValue):

    CHOICES = [
        'Sets',
        'Characters',
        'Props',
        'Audios',
        'Storyboards',
        'Layout scenes'
    ]


class ShotsMultichoiceValue(flow.values.MultiChoiceValue):
    
    _sequence = flow.Parent(3)
    
    def choices(self):
        return self._sequence.shots.mapped_names()


class SequenceElementSelection(flow.Object):
    
    _sequence = flow.Parent(2)
    select_all_elements = flow.SessionParam(False).ui(editor='bool').watched()
    elements = flow.Param([], ShotElements).ui(label="Elements to request")
    select_all_shots = flow.SessionParam(False).ui(editor='bool').watched()
    shots = flow.Param([], ShotsMultichoiceValue)
    
    def revert_to_defaults(self):
        self.elements.revert_to_default()
        self.shots.revert_to_default()
    
    def compute_oids(self, include_dependencies=False, latest_dependencies=False):
        oids = []
        elements = self.elements.get()
        
        for shot_name in self.shots.get():
            shot = self._sequence.shots[shot_name]
            shot.request.selection.elements.set(elements)
            
            oids += shot.request.selection.compute_oids(include_dependencies, latest_dependencies)
        
        return oids
    
    def child_value_changed(self, child_value):
        if child_value is self.select_all_shots:
            if self.select_all_shots.get():
                self.shots.set(self.shots.choices())
            else:
                self.shots.set([])
        elif child_value is self.select_all_elements:
            if self.select_all_elements.get():
                self.elements.set(self.elements.choices())
            else:
                self.elements.set([])


class RequestSequence(RequestRevisions):

    _sequence = flow.Parent()
    selection = flow.Child(SequenceElementSelection).ui(expanded=True)
    dependency_options = flow.Child(DependencyOptions).ui(expanded=True)
    revisions = flow.Child(RequestedRevisions).ui(expanded=True)
    pattern = flow.SessionParam('').ui(
        placeholder='Revision oid pattern',
        hidden=True,
    )
    
    def get_buttons(self):
        self.message.set('<h2>Request %s</h2>' % self._sequence.name())
        self.selection.revert_to_defaults()

        return super(RequestSequence, self).get_buttons()
    
    def run(self, button):
        if button == 'Close':
            return
        elif button == 'Request':
            return super(RequestSequence, self).run(button)
        
        oids = self.selection.compute_oids(
            include_dependencies=self.dependency_options.request_dependencies.get(),
            latest_dependencies=self.dependency_options.include_last_revisions.get()
        )
        self.pattern.set(';'.join(oids))
        
        return super(RequestSequence, self).run(button)


class MultiRequestSequence(MultiRequestRevisions):

    _sequence = flow.Parent()
    selection = flow.Child(SequenceElementSelection).ui(expanded=True)
    dependency_options = flow.Child(DependencyOptions).ui(expanded=True)
    revisions = flow.Child(MultiRequestedRevisions).ui(expanded=True)
    pattern = flow.SessionParam('').ui(
        placeholder='Revision oid pattern',
        hidden=True)
    
    def get_buttons(self):
        self.message.set('<h2>Request %s towards multiple sites</h2>' % self._sequence.name())
        self.selection.revert_to_defaults()

        return super(MultiRequestSequence, self).get_buttons()
    
    def run(self, button):
        if button == 'Close':
            return
        elif button == 'Request':
            return super(MultiRequestSequence, self).run(button)
        
        oids = self.selection.compute_oids(
            include_dependencies=self.dependency_options.request_dependencies.get(),
            latest_dependencies=self.dependency_options.include_last_revisions.get()
        )
        self.pattern.set(';'.join(oids))
        
        return super(MultiRequestSequence, self).run(button)


class PlayblastItem(flow.Object):

    to_submit = flow.SessionParam(False).ui(editor='bool').watched()
    use_simplify = flow.SessionParam(False).ui(editor='bool')
    reduce_textures = flow.SessionParam(False).ui(editor='bool')
    target_texture_width = flow.SessionParam(False).ui(editor='bool')
    priority = flow.SessionParam(10).ui(editor='int').watched()

    def child_value_changed(self, child_value):
        # Do nothing by default
        pass


class PlayblastFileRevision(flow.values.ChoiceValue):

    STRICT_CHOICES = False

    _file_pb = flow.Parent()

    def set_default_value(self, value):
        if value == 'last':
            revision_names = self._file_pb.revision_names()
            if revision_names:
                value = revision_names[-1]
            else:
                value = ''
        
        super(PlayblastFileRevision, self).set_default_value(value)

    def choices(self):
        return self._file_pb.revision_names() or []


class PlayblastFile(PlayblastItem):
    
    _task_pb = flow.Parent(2)
    _shot_pb = flow.Parent(4)
    _sequence = flow.Parent(7)

    revision = flow.Param('last', PlayblastFileRevision).watched()
    
    file_oid = flow.Computed(cached=True)
    display_name = flow.Computed(cached=True)
    department = flow.Computed(cached=True)
    exists = flow.Computed(cached=True)
    revision_available = flow.Computed(cached=True)

    def revision_names(self):
        if not self.exists.get():
            return None
        
        file_object = self.root().get_object(self.file_oid.get())
        return file_object.get_revision_names(published_only=True)

    def _compute_file_oid(self):
        return '%s/shots/%s/departments/%s/files/%s' % (
            self._sequence.oid(),
            self._shot_pb.name(),
            self.department.get(),
            self.name()
        )
    
    def _compute_exists(self):
        return self.root().session().cmds.Flow.exists(self.file_oid.get())
    
    def _compute_department(self):
        file_data = self._task_pb.file_data(self.name())
        return file_data['department']
    
    def _compute_display_name(self):
        file_data = self._task_pb.file_data(self.name())
        return file_data['display_name']
    
    def _compute_revision_available(self):
        if not self.exists.get() or not self.revision.get():
            return False
        
        file_object = self.root().get_object(self.file_oid.get())
        rev_object = file_object.get_revision(self.revision.get())

        return rev_object.get_sync_status() == 'Available'

    def compute_child_value(self, child_value):
        if child_value is self.file_oid:
            self.file_oid.set(self._compute_file_oid())
        elif child_value is self.exists:
            self.exists.set(self._compute_exists())
        elif child_value is self.display_name:
            self.display_name.set(self._compute_display_name())
        elif child_value is self.department:
            self.department.set(self._compute_department())
        elif child_value is self.revision_available:
            self.revision_available.set(self._compute_revision_available())
    
    def child_value_changed(self, child_value):
        if child_value is self.revision:
            self.revision_available.touch()
    
    def submit(self, pool_name):
        if not self.to_submit.get() or not self.exists.get() or not self.revision_available.get():
            return
        
        file_object = self.root().get_object(self.file_oid.get())

        file_object.submit_blender_playblast_job.revision_name.set(self.revision.get())
        file_object.submit_blender_playblast_job.use_simplify.set(self.use_simplify.get())
        file_object.submit_blender_playblast_job.reduce_textures.set(self.reduce_textures.get())
        file_object.submit_blender_playblast_job.target_texture_width.set(self.target_texture_width.get())
        file_object.submit_blender_playblast_job.priority.set(self.priority.get())
        file_object.submit_blender_playblast_job.pool.set(pool_name)
        file_object.submit_blender_playblast_job.run('Add to render pool')


class PlayblastFiles(flow.DynamicMap):
    
    _task_pb = flow.Parent()
    
    @classmethod
    def mapped_type(cls):
        return PlayblastFile
    
    def mapped_names(self, page_num=0, page_size=None):
        return self._task_pb.file_names()


class PlayblastTask(PlayblastItem):
    
    _shot_pb = flow.Parent(2)
    files = flow.Child(PlayblastFiles)

    def file_names(self):
        task_data = self._shot_pb.task_data(self.name())
        return task_data['files_data'].keys()

    def file_data(self, file_name):
        task_data = self._shot_pb.task_data(self.name())
        return task_data['files_data'][file_name]
    
    def child_value_changed(self, child_value):
        if child_value is self.to_submit:
            for file in self.files.mapped_items():
                file.to_submit.set(self.to_submit.get())
        elif child_value is self.priority:
            for file in self.files.mapped_items():
                file.priority.set(self.priority.get())
    
    def submit(self, pool_name):
        for file in self.files.mapped_items():
            file.submit(pool_name)


class PlayblastTasks(flow.DynamicMap):
    
    _shot_pb = flow.Parent()
    
    @classmethod
    def mapped_type(cls):
        return PlayblastTask
    
    def mapped_names(self, page_num=0, page_size=None):
        return self._shot_pb.task_names()


class PlayblastShot(PlayblastItem):
    
    _render_action = flow.Parent(2)
    tasks = flow.Child(PlayblastTasks)

    def task_names(self):
        return self._render_action.shot_task_names()
    
    def task_data(self, task_name):
        return self._render_action.shot_task_data(task_name)
    
    def child_value_changed(self, child_value):
        if child_value is self.to_submit:
            for task in self.tasks.mapped_items():
                task.to_submit.set(self.to_submit.get())
        elif child_value is self.priority:
            for task in self.tasks.mapped_items():
                task.priority.set(self.priority.get())
    
    def submit(self, pool_name):
        for task in self.tasks.mapped_items():
            task.submit(pool_name)


class PlayblastShots(flow.DynamicMap):
    
    sequence = flow.Parent(2)
    
    @classmethod
    def mapped_type(cls):
        return PlayblastShot
    
    def mapped_names(self, page_num=0, page_size=None):
        return self.sequence.shots.mapped_names()


class RenderSequencePlayblasts_XX(flow.Action):
    
    shots = flow.Child(PlayblastShots).ui(
        expanded=True,
        # action_submenus=True,
        items_action_submenus=True,
    )
    submit_all = flow.SessionParam(False).ui(editor='bool').watched()
    default_priority = flow.SessionParam(10).ui(editor='int').watched()
    pool = flow.Param('default', SiteJobsPoolNames)
    
    def __init__(self, parent, name):
        super(RenderSequencePlayblasts, self).__init__(parent, name)
        self._task_names = None
        self._tasks_data = None
    
    def _get_tasks_infos(self):
        template = self.root().project().admin.dependency_templates['shot']
        deps = template.get_dependencies()
        default_dep_names = template.get_default_dependency_names()
        
        task_names = []
        tasks_data = {}
        for name in default_dep_names:
            dep = deps[name]
            files_data = {}
            
            for file_name, file_data in dep['files'].items():
                if fnmatch(file_name, '*.blend'):
                    # Keep only Blender files
                    mapped_name = file_name.replace('.', '_')
                    files_data[mapped_name] = dict(
                        display_name=file_name,
                        department=file_data['department']
                    )
            
            if files_data:
                task_names.append(name)
                tasks_data[name] = dict(files_data=files_data)
        
        return task_names, tasks_data
    
    def _ensure_tasks_infos(self):
        if self._task_names is None or self._tasks_data is None:
            self._task_names, self._tasks_data = self._get_tasks_infos()
        
        return self._task_names, self._tasks_data
    
    def shot_task_names(self):
        return self._ensure_tasks_infos()[0]
    
    def shot_task_data(self, task_name):
        return self._ensure_tasks_infos()[1][task_name]
    
    def set_to_submit(self, to_submit):
        for shot in self.shots.mapped_items():
            shot.to_submit.set(to_submit)
    
    def set_default_priority(self, priority):
        for shot in self.shots.mapped_items():
            shot.priority.set(priority)
    
    def child_value_changed(self, child_value):
        if child_value is self.submit_all:
            self.set_to_submit(self.submit_all.get())
        elif child_value is self.default_priority:
            self.set_default_priority(self.default_priority.get())
    
    def get_buttons(self):
        return ['Submit playblasts', 'Cancel']
    
    def run(self, button):
        if button == 'Cancel':
            return
        
        for shot in self.shots.mapped_items():
            shot.submit(self.pool.get())


class ProcessSequenceFiles(flow.Action):

    _sequence = flow.Parent()

    def __init__(self, parent, name):
        super(ProcessSequenceFiles, self).__init__(parent, name)
        self._files_data = None
        self._task_names = []
    
    def ensure_files_data(self):
        '''
        Method used to cache computed files data
        '''
        if self._files_data is None:
            self._files_data = self._compute_files_data()
        
        return self._files_data
    
    def _compute_files_data(self):
        '''
        Computes data related to files contained in a sequence
        '''
        raise NotImplementedError(
            'Subclasses must return shot files data'
        )
    
    def get_shot_task_files(self, fnpatterns, tags=[]):
        self._task_names.clear()
        
        template = self.root().project().admin.dependency_templates['shot']
        default_dep_names = template.get_default_dependency_names()
        dependencies = template.get_dependencies()
        
        template_tasks_files = []
        
        for dep_name in default_dep_names:
            dep = dependencies[dep_name]
            files = []
            
            for file_name, file_data in dep['files'].items():
                file_tags = file_data.get('tags', [])
                
                if not file_tags and tags:
                    continue
                if not set(tags).issubset(set(file_tags)):
                    continue
                        
                for pattern in fnpatterns:
                    if fnmatch(file_name, pattern):
                        files.append((file_name, file_data))
            
            if files:
                template_tasks_files.append((
                    dep_name,
                    files
                ))
                self._task_names.append(dep_name)
        
        return template_tasks_files
    
    def get_shot_task_names(self):
        self.ensure_files_data()
        return self._task_names


class RenderSequencePlayblasts(ProcessSequenceFiles):
    
    ICON = ('icons.gui', 'film-strip-with-two-photograms')
    
    def get_job_pool_names(self):
        site = self.root().project().get_current_site()
        return ['default'] + site.pool_names.get()
    
    def submit_blender_playblast_job(self, shot_name, dept_name, file_name, revision_name, use_simplify, reduce_textures, target_texture_width, priority=10, pool_name='default'):
        file_name = file_name.replace('.', '_')
        oid = f'{self._sequence.oid()}/shots/{shot_name}/departments/{dept_name}/files/{file_name}'
        file = self.root().get_object(oid)
        
        file.submit_blender_playblast_job.revision_name.set(revision_name)
        file.submit_blender_playblast_job.use_simplify.set(use_simplify)
        file.submit_blender_playblast_job.reduce_textures.set(reduce_textures)
        file.submit_blender_playblast_job.target_texture_width.set(target_texture_width)
        file.submit_blender_playblast_job.priority.set(priority)
        file.submit_blender_playblast_job.pool.set(pool_name)
        file.submit_blender_playblast_job.run('Submit job')
    
    def submit_afterfx_playblast_job(self, shot_name, dept_name, file_name, revision_name, priority=10, pool_name='default'):
        file_name = file_name.replace('.', '_')
        oid = f'{self._sequence.oid()}/shots/{shot_name}/departments/{dept_name}/files/{file_name}'
        file = self.root().get_object(oid)
        
        file.submit_ae_playblast.revision.set(revision_name)
        # file.submit_ae_playblast.priority.set(priority)
        file.submit_ae_playblast.pool.set(pool_name)
        file.submit_ae_playblast.run('Submit')
    
    def _compute_files_data(self):
        # Keep only tasks which contain Blender files
        blender_files_by_task = self.get_shot_task_files(['*.blend', '*.aep'], ['playblast'])
        
        shots = self._sequence.shots
        files_data = []
        
        for shot in shots.mapped_items():
            tasks_files = []
            
            for task_name, task_files in blender_files_by_task:
                files = []
                
                for file_name, file_data in task_files:
                    file_mapped_name = file_name.replace('.', '_')
                    dept_name = file_data['department']
                    file_oid = f'{shot.oid()}/departments/{dept_name}/files/{file_mapped_name}'
                    
                    if not self.root().session().cmds.Flow.exists(file_oid):
                        # Skip if file does not exist in the flow
                        continue
                    
                    file = self.root().get_object(file_oid)
                    revision_statuses = file.get_revision_statuses(published_only=True)
                    head = file.get_head_revision(sync_status='Available')
                    
                    default_revision = head.name() if head is not None else None
                    
                    if not revision_statuses:
                        self.root().session().log_warning(
                            f'File {file.oid()} has no published revision available'
                        )
                        continue
                    
                    files.append(dict(
                        oid=file.oid(),
                        name=file_name,
                        shot=shot.name(),
                        department=dept_name,
                        revisions=revision_statuses,
                        default_revision=default_revision
                    ))
                
                tasks_files.append(dict(
                    name=task_name,
                    files=files
                ))
            
            files_data.append(dict(
                oid=shot.oid(),
                name=shot.name(),
                tasks=tasks_files
            ))
        
        return files_data
    
    def _fill_ui(self, ui):
        ui['custom_page'] = 'libreflow.utils.ui.RenderSequencePlayblastsWidget'


class UploadSequencePlayblastsToKitsu(ProcessSequenceFiles):
    
    ICON = ('icons.libreflow', 'kitsu-upload')

    def allow_context(self, context):
        return (
            context
            and self.root().project().kitsu_config().configured.get()
        )
    
    def get_task_statutes(self):
        return self.root().project().kitsu_api().get_task_statutes()
    
    def get_kitsu_task_type(self, task_name):
        dependencies = self.root().project().admin.dependency_templates['shot'].get_dependencies()
        kitsu_data = dependencies[task_name].get('kitsu', None)
        
        if kitsu_data is None:
            return None
        
        return kitsu_data.get('task_type', 'UNDEFINED')
    
    def upload_playblast_to_kitsu(self, file_oid, revision_name, task_type, target_task_status, comment):
        file = self.root().get_object(file_oid)
        file.upload_playblast.revision_name.set(revision_name)
        file.upload_playblast.target_task_type.set(task_type)
        file.upload_playblast.target_task_status.set(target_task_status)
        file.upload_playblast.comment.set(comment)
        
        file.upload_playblast.run(None)
    
    def _compute_files_data(self):
        # Search for playblast files
        playblasts_files_by_task = self.get_shot_task_files(['*.mov'])
        # print(blender_files_by_task)
        
        shots = self._sequence.shots
        files_data = []
        
        for shot in shots.mapped_items():
            tasks_files = []
            
            for task_name, task_files in playblasts_files_by_task:
                files = []
                default_kitsu_task_type = self.get_kitsu_task_type(task_name)
                
                for file_name, file_data in task_files:
                    file_mapped_name = file_name.replace('.', '_')
                    dept_name = file_data['department']
                    file_oid = f'{shot.oid()}/departments/{dept_name}/files/{file_mapped_name}'
                    
                    if not self.root().session().cmds.Flow.exists(file_oid):
                        # Skip if file does not exist in the flow
                        continue
                    
                    file = self.root().get_object(file_oid)
                    revision_statuses = file.get_revision_statuses(published_only=True)
                    head = file.get_head_revision(sync_status='Available')
                    
                    default_revision = head.name() if head is not None else None
                    
                    if not revision_statuses:
                        self.root().session().log_warning(
                            f'File {file.oid()} has no published revision available'
                        )
                        continue
                    
                    task_types = file.upload_playblast.target_task_type.choices()
                    
                    if default_kitsu_task_type is not None:
                        file.upload_playblast.target_task_type.set(default_kitsu_task_type)
                        
                        if len(task_types) == 1:
                            # Site has no Kitsu priviledge, and single task type depends on file name
                            # => Change to Kitsu task type specified by dependency
                            task_types = [default_kitsu_task_type]
                    else:
                        file.upload_playblast.target_task_type.revert_to_default()
                    
                    files.append(dict(
                        oid=file.oid(),
                        name=file_name,
                        shot=shot.name(),
                        department=dept_name,
                        revisions=revision_statuses,
                        default_revision=default_revision,
                        task_types=task_types,
                        default_task_type=file.upload_playblast.target_task_type.get(),
                        current_task_status=file.upload_playblast.current_task_status.get(),
                        default_task_status=file.upload_playblast.target_task_status.get()
                    ))
                
                tasks_files.append(dict(
                    name=task_name,
                    files=files
                ))
            
            files_data.append(dict(
                oid=shot.oid(),
                name=shot.name(),
                tasks=tasks_files
            ))
        
        return files_data
    
    def _fill_ui(self, ui):
        ui['custom_page'] = 'libreflow.utils.ui.UploadSequencePlayblastsWidget'


class Sequence(KitsuSequence):

    ICON = ("icons.flow", "sequence")

    _map = flow.Parent()

    settings = flow.Child(ContextualView).ui(hidden=True)
    description = flow.Param("")
    toggle_bookmark = flow.Child(ToggleBookmarkAction)
    shots = flow.Child(Shots).ui(expanded=True)
    
    request = flow.Child(RequestSequence).ui(hidden=True)
    multi_request = flow.Child(MultiRequestSequence)
    render_playblasts = flow.Child(RenderSequencePlayblasts)
    upload_playblasts = flow.Child(UploadSequencePlayblastsToKitsu)

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(sequence=self.name())


class ClearSequencesAction(ClearMapAction):
    def run(self, button):
        for sequence in self._map.mapped_items():
            for shot in sequence.shots.mapped_items():
                shot.kitsu_settings.clear()

            sequence.shots.clear()
            sequence.kitsu_settings.clear()

        super(ClearSequencesAction, self).run(button)


class Sequences(ItemMap):

    ICON = ("icons.flow", "sequence")

    item_prefix = "s"

    create_sequence = flow.Child(CreateItemAction)
    update_kitsu_settings = flow.Child(UpdateItemsKitsuSettings)

    @classmethod
    def mapped_type(cls):
        return Sequence

    def columns(self):
        return ["Name"]

    def _fill_row_cells(self, row, item):
        row["Name"] = item.name()

    def get_default_contextual_edits(self, context_name):
        if context_name == "settings":
            return dict(file_category="PROD")





import sys
import time
import traceback
import argparse
import os

from kabaret.app.session import KabaretSession

from .jobs_actor import Jobs


class JobsNodeSession(KabaretSession):

    def __init__(self, pool_names, *args, **kwargs):
        self.jobs_actor = None
        self.pool_names = pool_names
        self.queues = []
        self._current_job = None
        super(JobsNodeSession, self).__init__(*args, **kwargs)

    def _create_actors(self):
        super(JobsNodeSession, self)._create_actors()
        self.jobs_actor = Jobs(self)

    # def do_flow_job(self, oid, interpretor=None):
    #     #print 'Faking exection 5sec for oid={!r}'.format(oid)
    #     #time.sleep(5)
    #     print 'Spawning Worker to execute on oid={!r}'.format(oid)
    #     try:
    #         #self.cmds.Flow.call(oid, 'execute', (), {})
    #         popen = self.cmds.Jobs.node_spawn_flow_worker()
    #     except:
    #         traceback.print_exc()
    #         print 'Oops :/'
    #         raise

    # def _process_job(self, job):
    #     print 'Got job:', job.to_dict()
    #     job_type = job.get_job_type()
    #     if job_type == 'NODE_CMD':
    #         params = job.get_job_params()
    #         print ' ---> node cmd:', params
    #         args, kwargs = params
    #         self._current_job = job
    #         try:
    #             self.do_node_control_job(*args, **kwargs)
    #         except:
    #             traceback.print_exc()
    #         else:
    #             print 'Command done'
    #             job.set_done()
    #         finally:
    #             self._current_job = None

    #     elif job_type == 'flow':
    #         params = job.get_job_params()
    #         print ' ---> Params:', params
    #         args, kwargs = params
    #         try:
    #             self.do_flow_job(*args, **kwargs)
    #         except:
    #             traceback.print_exc()
    #         else:
    #             print 'Marking job done'
    #             job.set_done()

    def wait_for_jobs(self):
        node_id = self.session_uid()
        working_job_popen = None

        self.log_info('Jobs Node {} started...'.format(self.session_uid()))

        while not self.jobs_actor.node_stopped():
            self.tick()

            self.log_debug('{}:: waiting for control jobs...'.format(self.session_uid()))
            control_job = self.jobs_actor.poll_control(node_id)
            if control_job is not None:
                #self._process_job(control_job)
                self.log_info('{}:: received control job {} with command {}'.format(
                    self.session_uid(),
                    control_job.jid(),
                    control_job.get_job_params()[1].get('cmd', None)))
                self.jobs_actor.node_execute_job(control_job)

            self.tick()

            if working_job_popen is not None:
                retcode = working_job_popen.poll()
                if retcode is not None:
                    # process ended
                    working_job_popen = None

            if working_job_popen is None:
                self.log_debug('{}:: waiting for jobs...'.format(self.session_uid()))
                job = self.jobs_actor.poll_pools(self.pool_names, node_id)
                if job is not None:
                    #self._process_job(job)
                    working_job_popen = self.jobs_actor.node_execute_job(job)
                    self.log_info('Jobs Node worker started. PID: {}'.format(working_job_popen.pid))
            
            else:
                self.log_debug('Jobs Node worker in progress. PID: {}'.format(working_job_popen.pid))

            time.sleep(1)


def process_remaining_args(args):
    parser = argparse.ArgumentParser(
        description='Libreflow Session Arguments'
    )
    parser.add_argument(
        '-r', '--root_dir', dest='root_dir'
    )
    values, remaining_args = parser.parse_known_args(args)

    if values.root_dir:
        os.environ["ROOT_DIR"] = values.root_dir
    
    return remaining_args


if __name__ == '__main__':
    argv = sys.argv[1:]  # get ride of first args wich is script filename
    session_name, host, port, cluster_name, db, password, debug, remaining_args \
        = JobsNodeSession.parse_command_line_args(argv)
    pools = process_remaining_args(remaining_args)
    if not pools:
        print('\n!!!\nUsage:', sys.argv[0], 'pool_name pool_name ...\n\n')
    else:
        node = JobsNodeSession(pools, session_name=session_name, debug=debug)
        node.cmds.Cluster.connect(host, port, cluster_name, db, password)

        node.wait_for_jobs()
        print('Jobs Node closed.')
        



import os
import sys
import gazu
import time
from datetime import datetime, timedelta
import subprocess
import platform
from minio import Minio
import zipfile
import fnmatch
import re
import traceback
import tempfile
import uuid
import pathlib

from kabaret import flow
from kabaret.flow_contextual_dict import ContextualView, get_contextual_dict
from kabaret.subprocess_manager.flow import RunAction
from kabaret.flow_entities.entities import EntityCollection, Entity, Property, PropertyValue

from ..utils.os import zip_folder, unzip_archive

from .dependency import get_dependencies
from .maputils import ItemMap, CreateGenericAction, ClearMapAction, RemoveGenericAction, SimpleCreateAction
from .runners import Runner, PythonRunner


class StaticSiteTypeChoices(flow.values.ChoiceValue):
    CHOICES = ['Studio', 'User']


class StaticSiteSyncStatusChoices(flow.values.ChoiceValue):
    CHOICES = ['NotAvailable', 'Requested', 'Available']


class StaticSiteSyncTransferStateChoices(flow.values.ChoiceValue):
    CHOICES = ['NC', 'UP_REQUESTED', 'DL_REQUESTED', 'UP_IN_PROGRESS', 'DL_IN_PROGRESS']


class LoadType(PropertyValue):

    DEFAULT_EDITOR = 'choice'
    CHOICES = ['Upload', 'Download']

    def choices(self):
        return self.CHOICES


class JobStatus(PropertyValue):

    DEFAULT_EDITOR = 'choice'
    CHOICES = ['WFA', 'WAITING', 'PROCESSED', 'ERROR', 'PAUSE', 'DONE']

    def choices(self):
        return self.CHOICES


class ResetJob(flow.Action):

    ICON = ('icons.libreflow', 'reset')

    _job = flow.Parent()
    _jobs = flow.Parent(2)

    def needs_dialog(self):
        o = self.root().get_object(self._job.emitter_oid.get())
        return self._job.type.get() == 'Upload' and o.get_sync_status(exchange=True) == 'Available'
    
    def allow_context(self, context):
        return context and self._job.status.get() == 'ERROR'
    
    def get_buttons(self):
        self.message.set((
            '<h3>Revision already on the exchange server</h3>'
            'Reset upload job anyway ?'
        ))
        return ['Confirm', 'Cancel']

    def run(self, button):
        if button == 'Cancel':
            return
        
        self._job.status.set('WAITING')
        self._job.log.set('?')
        self._job.on_submitted()
        self._jobs.incr_waiting_count()
        o = self.root().get_object(self._job.emitter_oid.get())
        o.set_sync_status('NotAvailable', exchange=True)


class ResetJobs(flow.Action):

    ICON = ('icons.libreflow', 'reset')

    _jobs = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        if button == 'Cancel':
            return
        
        for j in self._jobs.jobs(status='ERROR'):
            j.status.set('WAITING')
        
        self._jobs.touch()


class ShowRevisionInHistory(flow.Action):

    ICON = ("icons.gui", "ui-layout")

    _job = flow.Parent()

    def needs_dialog(self):
        return False

    def run(self, button):
        history_oid = self._job.emitter_oid.get() + '/../..'
        history_oid = self.root().session().cmds.Flow.resolve_path(history_oid)

        return self.get_result(goto=history_oid)


class Job(Entity):

    type                = Property(property_value_type=LoadType)
    status              = Property(property_value_type=JobStatus)
    priority            = Property().ui(editor='int')
    emitter_oid         = Property()
    path                = Property()
    date                = Property().ui(editor='datetime')
    date_end            = Property().ui(editor='datetime')
    is_archived         = Property().ui(editor='bool')
    requested_by_user   = Property()
    requested_by_studio = Property()
    log                 = Property().ui(editor='textarea')

    reset = flow.Child(ResetJob)
    show = flow.Child(ShowRevisionInHistory).ui(label='Show in history')

    def __repr__(self):
        job_repr = "%s(type=%s, status=%s, priority=%s, emitter=%s, date=%s, date_end=%s, is_archived=%s, requested_by_user=%s, requested_by_studio=%s)" % (
                self.__class__.__name__,
                self.type.get(),
                self.status.get(),
                self.priority.get(),
                self.emitter_oid.get(),
                self.date.get(),
                self.date_end.get(),
                self.is_archived.get(),
                self.requested_by_user.get(),
                self.requested_by_studio.get()
            )

        return job_repr
    
    def get_local_path(self):
        '''
        Returns the local path of the associated object.
        
        Note that the result depends on the current site
        since it includes the root folder of the latter.
        '''
        return os.path.join(
            self.root().project().get_root(),
            self.path.get()
        )
    
    def get_server_path(self):
        '''
        Returns the path of the associated object on the server.
        '''
        path = self.path.get()

        if os.path.splitext(path)[1] == '':
            path += '/' + os.path.basename(path) + '.zip'
        
        return path
    
    def get_related_object(self):
        try:
            o = self.root().get_object(
                self.emitter_oid.get()
            )
        except (ValueError, flow.exceptions.MissingRelationError):
            o = None
        
        return o
    
    def check_valid_state(self):
        '''
        Returns a boolean meaning whether this job is ready
        to be processed or not.
        
        The method may update the job's properties to indicate
        why this job cannot yet be processed. It assumes that
        `emitter_oid` references a valid revision object.
        '''
        if self.status.get() != 'WAITING':
            return False
        
        o = self.get_related_object()

        if self.type.get() == 'Download' and o.get_sync_status(exchange=True) != 'Available':
            self.log.set((
                'Revision is not yet available '
                'on the exchange server'
            ))
            return False
        
        return True
    
    def on_submitted(self):
        '''
        Method which may be called right after this job
        has been submitted to a queue.

        Its default behaviour is to compute the relative path
        of the job's requested revision (`emitter_oid`).
        If the oid does not refer to an existing revision
        object, the methods puts the job in error state.
        '''
        revision = self.get_related_object()
        
        if revision is None:
            self.status.set('ERROR')
            self.log.set('Object does not exist')
            return
        
        try:
            ready_for_sync = revision.ready_for_sync.get()
            path = revision.get_path(relative=True).replace('\\', '/')
        except AttributeError:
            self.status.set('ERROR')
            self.log.set('Object is not a revision')
            return
        else:
            if not ready_for_sync:
                self.status.set('ERROR')
                self.log.set('Revision is not ready to be synced')
                return
        
        self.path.set(path)
    
    def on_processed(self):
        '''
        Method which may be called right after this job
        has been processed.

        Its default behaviour is to update the related
        revision statutes. The method assumes that
        `emitter_oid` references a valid revision object.
        '''
        if self.status.get() != 'PROCESSED':
            return
        
        revision = self.root().get_object(
            self.emitter_oid.get()
        )
        revision.set_sync_status(
            'Available',
            exchange=(self.type.get() == 'Upload')
        )
        revision.touch()
        revision._revisions.touch()


class JobQueue(EntityCollection):

    count_waiting = flow.IntParam(0)
    reset_jobs = flow.Child(ResetJobs).ui(label='Reset erroneous jobs')

    @classmethod
    def mapped_type(cls):
        return Job
    
    def get_next_waiting_job(self):
        for job in reversed(self.mapped_items()):
            if job.status.get() == "WAITING":
                return job
        
        return None
    
    def jobs(self, type=None, status=None):
        jobs = self.mapped_items()

        if type is not None:
            jobs = [j for j in jobs if j.type.get() == type]
        if status is not None:
            jobs = [j for j in jobs if j.status.get() == status]
        
        return jobs
    
    def count(self, type=None, status=None):
        job_filter = {}

        if type is not None:
            job_filter['type'] = type
        if status is not None:
            job_filter['status'] = status
        
        c = self.get_entity_store().get_collection(self.collection_name())

        return c.count_documents(job_filter)
    
    def update_waiting_count(self, count=None):
        if count is None:
            count = len(self.jobs(status='WAITING'))
        
        self.count_waiting.set(count)
    
    def incr_waiting_count(self, by=1):
        self.count_waiting.incr(by)

    def decr_waiting_count(self, by=1):
        self.count_waiting.decr(by)
    
    def submit_job(self,
            emitter_oid,
            user,
            studio,
            date_end=-1,
            job_type='Download',
            init_status='WAITING',
            priority=50
        ):
        
        name = '%s_%s_%s_%i' % (
            emitter_oid[1:].replace('/', '_'),
            studio,
            job_type,
            time.time()
        )
        self.ensure_exist([name])
        self._document_cache = None

        job = self.get_mapped(name)
        job.type.set(job_type)
        job.status.set(init_status)
        job.priority.set(priority)
        job.emitter_oid.set(emitter_oid)
        job.date.set(time.time())
        job.date_end.set(date_end)
        job.requested_by_user.set(user)
        job.requested_by_studio.set(studio)
        job.is_archived.set(False)
        job.log.set('?')

        job.on_submitted()

        self.incr_waiting_count()

        self.root().session().log_info("Submitted job %s" % job)

        self.touch()

        return job
    
    def columns(self):
        return [
            "Type",
            "Status",
            "Revision",
            "Emitted on",
            "By",
            "Site",
        ]
    
    def _fill_row_cells(self, row, item):
        self.mapped_names()
        item_data = self._document_cache[item.name()]

        row['Type']       = item_data['type']
        row['Status']     = item_data['status']
        row['Revision']   = item_data['emitter_oid']
        row['By']         = item_data['requested_by_user']
        row['Site']       = item_data['requested_by_studio']
        row['Emitted on'] = datetime.fromtimestamp(
            float(item_data['date'])
        ).ctime()


class ProcessJobs(flow.Action):

    def __init__(self, parent, name):
        super(ProcessJobs, self).__init__(parent, name)
        self._job_count = {}

    def needs_dialog(self):
        return False

    def process(self, job):
        raise NotImplementedError(
            "Must be implemented to process the given job"
        )
    
    def _get_jobs(self):
        current_site = self.root().project().get_current_site()
        return current_site.get_jobs()
    
    def _compute_job_count(self):
        count = dict.fromkeys(JobStatus.CHOICES, 0)

        for j in self.root().project().get_current_site().get_jobs():
            count[j.status.get()] += 1
        
        return count
    
    def run(self, button):
        self._job_count = self._compute_job_count()

        for job in self._get_jobs():
            self.root().session().log_info("Processing job %s" % job)
            
            self._job_count[job.status.get()] -= 1
            self.process(job)
            self._job_count[job.status.get()] += 1
        
        current_site = self.root().project().get_current_site()
        current_site.queue.job_list.update_waiting_count(self._job_count['WAITING'])

        # Refresh project's sync section UI
        self.root().project().synchronization.touch()


class MinioFileUploader(PythonRunner):
    
    def argv(self):
        args = ["%s/../scripts/minio_file_uploader.py" % (
            os.path.dirname(__file__)
        )]
        args += self.extra_argv
        return args


class MinioFileDownloader(PythonRunner):
    
    def argv(self):
        args = ["%s/../scripts/minio_file_downloader.py" % (
            os.path.dirname(__file__)
        )]
        args += self.extra_argv
        return args


class MinioUploadFile(flow.Object):

    def upload(self, local_path, server_path):
        self.root().session().log_info(
            "Uploading file %s -> %s" % (
                local_path,
                server_path
            )
        )
        exchange_site = self.root().project().get_exchange_site()
        minioClient = Minio(
            exchange_site.server_url.get(),
            access_key=exchange_site.server_login.get(),
            secret_key=exchange_site.server_password.get(),
            secure=True
        )

        minioClient.fput_object(
            exchange_site.bucket_name.get(),
            server_path,
            local_path
        )


class MinioDownloadFile(flow.Object):

    def download(self, server_path, local_path):
        self.root().session().log_info(
            "Downloading file %s -> %s" % (
                server_path,
                local_path
            )
        )
        exchange_site = self.root().project().get_exchange_site()
        minioClient = Minio(
            exchange_site.server_url.get(),
            access_key=exchange_site.server_login.get(),
            secret_key=exchange_site.server_password.get(),
            secure=True
        )
        
        tmp_path = self.root().project().get_temp_folder()

        if tmp_path is not None:
            tmp_name = os.path.splitext(os.path.basename(local_path))[0]
            tmp_path = os.path.join(tmp_path, tmp_name)

        minioClient.fget_object(
            exchange_site.bucket_name.get(),
            server_path,
            local_path,
            tmp_file_path=tmp_path
        )


class SyncManager(flow.Object):

    _exchange = flow.Parent()

    def __init__(self, parent, name):
        super(SyncManager, self).__init__(parent, name)
        self._client = None

    def upload(self, server_path, local_path):
        zipped_folder = self._is_zipped_folder(
            server_path, local_path
        )
        if zipped_folder:
            # Folder to be zipped before upload
            src_path = self._get_temp_zip(server_path)
            zip_folder(local_path, src_path)
        else:
            # File/folder uploaded as is
            src_path = local_path

        self.root().session().log_info(
            f'Upload file {local_path} -> {server_path}'
        )
        self._ensure_client().fput_object(
            self._exchange.bucket_name.get(),
            server_path,
            src_path
        )

        if zipped_folder:
            os.remove(src_path)

    def download(self, server_path, local_path):
        zipped_folder = self._is_zipped_folder(
            server_path, local_path
        )
        if zipped_folder:
            dst_path = self._get_temp_zip(server_path)
        else:
            dst_path = local_path
        
        self.root().session().log_info(
            f'Download file {server_path} -> {local_path}'
        )
        self._ensure_client().fget_object(
            self._exchange.bucket_name.get(),
            server_path,
            dst_path
        )

        if zipped_folder:
            unzip_archive(dst_path, local_path)
            os.remove(dst_path)
    
    def check_connection(self):
        try:
            self._ensure_client().list_buckets()
        except Exception as err:
            return str(err)
        else:
            return None

    def _ensure_client(self):
        if self._client is None:
            self._client = Minio(
                self._exchange.server_url.get(),
                access_key=self._exchange.server_login.get(),
                secret_key=self._exchange.server_password.get(),
                secure=self._exchange.enable_tls.get()
            )
        
        return self._client
    
    def _get_temp_zip(self, server_path):
        return os.path.join(
            self.root().project().get_temp_folder(),
            (
                pathlib.PurePath(server_path).stem
                + f'-{str(uuid.uuid4())}.zip'
            )
        )
    
    def _is_zipped_folder(self, server_path, local_path):
        return (
            os.path.splitext(local_path)[1] == ''
            and os.path.splitext(server_path)[1] == '.zip'
        )


class Synchronize(ProcessJobs):

    def _get_jobs(self):
        return self.root().project().get_current_site().get_jobs(
            status='WAITING'
        )

    def process(self, job):
        if not job.check_valid_state():
            return
        
        try:
            sync_manager = self.root().project().get_exchange_site().sync_manager
        
            if job.type.get() == 'Upload':
                sync_manager.upload(
                    job.get_server_path(),
                    job.get_local_path()
                )
            else:
                sync_manager.download(
                    job.get_server_path(),
                    job.get_local_path()
                )
        except Exception as e:
            job.status.set('ERROR')
            job.log.set(traceback.format_exc())
        else:
            job.status.set('PROCESSED')
            job.on_processed()


class ActiveSiteChoiceValue(flow.values.SessionValue):
    
    DEFAULT_EDITOR = 'choice'
    
    _choices = flow.SessionParam(None).ui(editor='set')

    def choices(self):
        if self._choices.get() is None:
            working_sites = self.root().project().get_working_sites()
            names = working_sites.get_site_names(use_custom_order=True, active_only=True)
            self._choices.set(names)
        
        return self._choices.get()


class ActiveSiteAutoSelectChoiceValue(ActiveSiteChoiceValue):
    
    def choices(self):
        choices = super(ActiveSiteAutoSelectChoiceValue, self).choices()
        choices = ['Auto select'] + choices.copy()
        
        return choices


class ActiveSitesMultichoiceValue(ActiveSiteChoiceValue):
    
    DEFAULT_EDITOR = 'multichoice'
    
    exclude_choice = flow.SessionParam()
    
    def choices(self):
        if self._choices.get() is None:
            working_sites = self.root().project().get_working_sites()
            names = working_sites.get_site_names(use_custom_order=True, active_only=True)
            self._choices.set(names)
        
        choices = self._choices.get().copy()
        exclude_choice = self.exclude_choice.get()
        
        if exclude_choice is not None:
            try:
                choices.remove(exclude_choice)
            except ValueError:
                pass
        
        return choices


class ActiveSiteRevisionAvailableChoiceValue(flow.values.ChoiceValue):
    
    DEFAULT_EDITOR = 'choice'
    
    _choices = flow.SessionParam(None).ui(editor='set')
    _revision = flow.Parent(2)
    
    exclude_choice = flow.SessionParam(None).ui(hidden=False)

    def choices(self):
        if self._choices.get() is None:
            working_sites = self.root().project().get_working_sites()
            choices = working_sites.get_site_names(use_custom_order=True, active_only=True)
            exclude_choice = self.exclude_choice.get()
            
            if exclude_choice is not None:
                try:
                    choices.remove(exclude_choice)
                except ValueError:
                    pass
            
            choices_available = []
            
            for site_name in choices:
                if self._revision.get_sync_status(site_name) == 'Available':
                    choices_available.append(site_name)
            
            self._choices.set(choices_available)
        
        return self._choices.get()


class SiteSelection(flow.Object):
    
    source_site = flow.Param(None, ActiveSiteChoiceValue).watched()
    target_site = flow.Param(None, ActiveSiteChoiceValue).watched()
    auto_select_enabled = flow.SessionParam(False).watched().ui(
        editor='bool',
        hidden=True)
    
    def child_value_changed(self, child_value):
        if child_value in (self.source_site, self.target_site, self.auto_select_enabled):
            self.touch()
    
    def summary(self):
        if self.source_site.get() == self.target_site.get() and not self.auto_select_enabled.get():
            return '⚠️ Source and target sites can\'t be the same !'
        else:
            return ''


class RequestAs(flow.Action):

    _revision = flow.Parent()
    sites = flow.Child(SiteSelection).ui(expanded=True)
    priority = flow.IntParam(50)
    include_dependencies = flow.SessionParam(True).ui(editor='bool')
    forced_upload = flow.SessionParam(False).ui(
        editor='bool',
        tooltip=('Ask the source site to upload the revision regardless'
                 ' of its status on the exchange site.'))
    
    def get_buttons(self):
        self.sites.source_site.set(self.root().project().get_current_site().name())
        target_site_choices = self.sites.target_site.choices()
        
        if not target_site_choices:
            msg = self.message.get()
            if msg is None:
                msg = ''
            
            msg += (
                '<h3><font color="#D5000D">Making requests is not possible since '
                'there is no other site defined for this project.</font></h3>'
            )
            self.message.set(msg)
            
            return ['Cancel']
        
        self.sites.target_site.set(target_site_choices[0])

        return ["Request", "Cancel"]
    
    def get_source_site_name(self):
        return self.sites.source_site.get()
    
    def get_target_site_name(self):
        return self.sites.target_site.get()
    
    def allow_context(self, context):
        return (
            context
            and not self._revision.is_working_copy()
            and self.root().project().get_current_site().request_files_from_anywhere.get()
        )

    def run(self, button):
        if button == "Cancel":
            return
        
        source_site_name = self.get_source_site_name()
        target_site_name = self.get_target_site_name()

        # Get requesting and requested sites
        sites = self.root().project().get_working_sites()
        target_site = sites[target_site_name]
        
        # Add a download job for the requesting site
        target_site.get_queue().submit_job(
            job_type="Download",
            init_status="WAITING",
            emitter_oid=self._revision.oid(),
            user=self.root().project().get_user_name(),
            studio=target_site_name,
            priority=self.priority.get(),
        )
        self._revision.set_sync_status("Requested", site_name=target_site_name)
        self._revision._revisions.touch()
        
        # Request dependencies if required
        if self.include_dependencies.get():
            request_deps = self._revision.request_dependencies
            request_deps.target_site.set(target_site_name)
            request_deps.predictive_only.set(False)
            request_deps.run(None)
        
        # Check if the version is not available on the exchange server
        exchange_site = self.root().project().get_exchange_site()
        revision_status = self._revision.get_sync_status(exchange=True)

        if revision_status == "Available" and not self.forced_upload.get():
            self.root().session().log_warning(
                "Revision already on the exchange server"
            )
            self.root().project().synchronization.touch()
            return self.get_result()
        
        # If enabled, automatically select source site
        if self.sites.auto_select_enabled.get():
            source_site_name = self._revision.site.get()
        
        source_site = sites[source_site_name]

        # Check if the source version upload is not already requested
        for job in source_site.get_jobs(type="Upload", status="WAITING"):
            if job.emitter_oid.get() == self._revision.oid():
                self.root().session().log_warning(
                    "Revision already requested for upload in source site"
                )
                self.root().project().synchronization.touch()
                return self.get_result()
        
        # Add an upload job for the requested site
        source_site.get_queue().submit_job(
            job_type="Upload",
            init_status="WAITING",
            emitter_oid=self._revision.oid(),
            user=self.root().project().get_user_name(),
            studio=target_site_name,
            priority=self.priority.get(),
        )

        # Refresh project's sync section UI
        self.root().project().synchronization.touch()


class Request(RequestAs):

    ICON = ('icons.libreflow', 'request')

    _revision = flow.Parent()
    priority = flow.IntParam(50)
    source_site = flow.Param(None, ActiveSiteRevisionAvailableChoiceValue)
    
    sites = flow.Child(SiteSelection).ui(hidden=True)

    def get_buttons(self):
        self.source_site.exclude_choice.set(self.root().project().get_current_site().name())
        source_site_choices = self.source_site.choices()
        
        if not source_site_choices:
            msg = self.message.get()
            if msg is None:
                msg = ''
            
            msg += (
                '<h3><font color="#D5000D">Making requests is not possible since '
                'the revision isn\'t available on any site.</font></h3>'
            )
            self.message.set(msg)
            
            return ['Cancel']
        
        self.source_site.set(source_site_choices[0])

        return ['Request', 'Cancel']
    
    def get_source_site_name(self):
        return self.source_site.get()
    
    def get_target_site_name(self):
        return self.root().project().get_current_site().name()
    
    def allow_context(self, context):
        return (
            context
            and not self._revision.is_working_copy()
            and self._revision.get_sync_status() != "Available"
            and self._revision.get_sync_status(exchange=True) != "Available"
        )


class SyncErrorCounter(flow.DynamicMap):

    _error_count = flow.Param([])

    def refresh_error_count(self):
        count = []
        sites = self.root().project().get_working_sites()
        site_names = sites.get_site_names(use_custom_order=True, active_only=True)

        for name in site_names:
            s = sites[name]
            count.append((s.name(), s.oid(), len(s.get_jobs(status='ERROR'))))
        
        self._error_count.set(count)
    
    def columns(self):
        return ['Site', 'Nb errors']

    def rows(self):
        count = self._error_count.get()
        rows = []

        for site_name, site_oid, nb_err in count:
            style = {'icon': ('icons.libreflow', 'blank')}
            oid = site_oid + '/queue'
            
            if nb_err > 0:
                style['foreground-color'] = '#D5000D'
                style['Nb errors_foreground-color'] = '#D5000D'
            
            rows.append((oid,
                {
                    'Site': site_name,
                    'Nb errors': nb_err,
                    'activate_oid': oid,
                    '_style': style
                }
            ))
        
        return rows


class ShowSiteSyncErrors(flow.Action):

    error_counter = flow.Child(SyncErrorCounter).ui(
        label='Errors',
        expanded=True)

    def get_buttons(self):
        self.message.set('<h2>Synchronization errors</h2>')
        return ['Refresh', 'Close']
    
    def run(self, button):
        if button == 'Close':
            return
        
        self.error_counter.refresh_error_count()
        self.error_counter.touch()
        return self.get_result(close=False)


class ResetJobStatuses(flow.Action):
    _site = flow.Parent()
    status = flow.Param("WAITING", JobStatus)

    def run(self, button):
        for job in self._site.get_jobs():
            job.status.set(self.status.get())
        
        self._site.get_queue().touch()


class CreateSite(flow.Action):

    ICON = ("icons.gui", "plus-sign-in-a-black-circle")

    _site_map = flow.Parent()
    site_name = flow.SessionParam("").ui(label="Name")

    def get_buttons(self):
        self.message.set("<h2>Create a site</h2>")
        return ["Create", "Cancel"]
    
    def run(self, button):
        if button == "Cancel":
            return
        
        site_name = self.site_name.get()

        if not site_name or self._site_map.has_mapped_name(site_name):
            if not site_name:
                msg = "Site name must not be empty."
            else:
                msg = f"Site {site_name} already exists."

            self.message.set((
                "<h2>Create a site</h2>"
                f"<font color=#D5000D>{msg}</font>"
            ))
            
            return self.get_result(close=False)

        site = self._site_map.add(site_name)
        self._site_map.touch()

        return self.get_result(next_action=site.configuration.oid())


class ConfigureSite(flow.Action):

    _site_map = flow.Parent(2)
    _site = flow.Parent()
    short_name = flow.SessionParam("").ui(label="Short name")
    description = flow.SessionParam("")

    def get_buttons(self):
        self.message.set(
            "<h2>Site <font color=#fff>%s</font></h2>" % self._site.name()
        )
        self._fill_action_fields()

        return ["Configure", "Cancel"]

    def _configure_site(self, site):
        '''
        This can be used by subclass to configure a mapped site.

        Default is to set site's short name and description.
        '''
        self._site.short_name.set(self.short_name.get())
        self._site.description.set(self.description.get())
    
    def _fill_action_fields(self):
        '''
        This can be used by subclass to fill action's parameters when
        dialog is displayed (e.g. to automatically show site parameters).

        Default is to fill site's short name and description.
        '''
        self.short_name.set(self._site.short_name.get())
        self.description.set(self._site.description.get())
    
    def allow_context(self, context):
        # Hide in map item submenus
        return False
    
    def run(self, button):
        if button == "Cancel":
            return
        
        self._configure_site(self._site)
        self._site.configured.touch()
        self._site_map.touch()


class ComputedBoolValue(flow.values.ComputedValue):

    DEFAULT_EDITOR = 'bool'


class Site(flow.Object):

    short_name = flow.Param("")
    description = flow.Param("")
    # Define conditions required for a site to be considered as correctly configured
    configured = flow.Computed().ui(editor='bool')
    configuration = flow.Child(ConfigureSite)

    is_active = flow.BoolParam(True)

    def compute_child_value(self, child_value):
        if child_value is self.configured:
            self.configured.set(True)


class SiteMap(flow.Map):

    create_site = flow.Child(CreateSite)

    _short_names = flow.HashParam()
    _active_site_names = flow.OrderedStringSetParam()

    @classmethod
    def mapped_type(cls):
        return Site

    def mapped_names(self, page_num=0, page_size=None):
        return ["default"] + super(SiteMap, self).mapped_names(page_num, page_size)
    
    def columns(self):
        return ["Site"]
    
    def short_names(self):
        short_names = []
        
        for name in self.mapped_names():
            short_names.append(self.get_short_name(name))
        
        return short_names
    
    def get_short_name(self, site_name):
        return self._short_names.get_key(site_name)
    
    def update_active_site_names(self):
        self._active_site_names.revert_to_default()
        i = 0

        for s in self.mapped_items():
            if s.is_active.get():
                self._active_site_names.add(s.name(), i)
            
            i += 1
    
    def update_site_short_names(self):
        self._short_names.revert_to_default()

        for s in self.mapped_items():
            self._short_names.set_key(s.name(), s.short_name.get())
    
    def get_site_names(self, use_custom_order=False, active_only=False, short_names=False):
        sites_data = self.root().project().admin.multisites.sites_data.get()
        
        if use_custom_order:
            current_site = self.root().project().get_current_site()
            names = current_site.ordered_site_names.get()
        else:
            names = self.mapped_names()

        site_names = []

        for name in names:
            data = sites_data[name]

            if active_only and not data['is_active']:
                continue
            
            if short_names:
                site_names.append(data['short_name'])
            else:
                site_names.append(name)
        
        return site_names
    
    def _get_mapped_item_type(self, mapped_name):
        if mapped_name == "default":
            return self.mapped_type()

        return super(SiteMap, self)._get_mapped_item_type(mapped_name)
    
    def _fill_row_cells(self, row, item):
        row["Site"] = item.name()
        if not item.configured.get():
            row["Site"] += " ⚠️"


class ConfigureWorkingSite(ConfigureSite):

    site_type = flow.Param("Studio", StaticSiteTypeChoices)

    root_windows_folder = flow.SessionParam("")
    root_linux_folder = flow.SessionParam("")
    root_darwin_folder = flow.SessionParam("")
    
    sync_dl_max_connections = flow.SessionParam(1)
    sync_up_max_connections = flow.SessionParam(1)
    
    def _fill_action_fields(self):
        super(ConfigureWorkingSite, self)._fill_action_fields()
        self.site_type.set(self._site.site_type.get())
        self.root_windows_folder.set(self._site.root_windows_folder.get())
        self.root_linux_folder.set(self._site.root_linux_folder.get())
        self.root_darwin_folder.set(self._site.root_darwin_folder.get())
        self.sync_dl_max_connections.set(self._site.sync_dl_max_connections.get())
        self.sync_up_max_connections.set(self._site.sync_up_max_connections.get())

    def _configure_site(self, site):
        super(ConfigureWorkingSite, self)._configure_site(site)
        site.site_type.set(self.site_type.get())
        site.root_windows_folder.set(self.root_windows_folder.get())
        site.root_linux_folder.set(self.root_linux_folder.get())
        site.root_darwin_folder.set(self.root_darwin_folder.get())
        site.sync_dl_max_connections.set(self.sync_dl_max_connections.get())
        site.sync_up_max_connections.set(self.sync_up_max_connections.get())


class Queue(flow.Object):

    job_list = flow.Child(JobQueue).ui(
        label="Jobs",
        expanded=True,
        show_filter=True)


class GotoQueue(flow.Action):

    _site = flow.Parent()

    def needs_dialog(self):
        return False
    
    def run(self, button):
        return self.get_result(goto=self._site.queue.oid())


class GotoCurrentSiteQueue(flow.Action):
    
    ICON = ('icons.flow', 'jobs')
    
    def needs_dialog(self):
        return False
    
    def allow_context(self, context):
        return context and context.endswith('.details')
    
    def run(self, button):
        current_site = self.root().project().get_current_site()
        return self.get_result(goto=current_site.queue.oid())


class ClearQueueJobStatus(flow.values.ChoiceValue):

    CHOICES = JobStatus.CHOICES + ["All"]


class ClearQueueLoadType(flow.values.ChoiceValue):

    CHOICES = LoadType.CHOICES + ["All"]


class ClearQueueAction(flow.Action):

    _site = flow.Parent()
    status = flow.Param("PROCESSED", ClearQueueJobStatus)
    type = flow.Param("All", ClearQueueLoadType)
    emitted_before = flow.SessionParam((datetime.now() - timedelta(days=1)).timestamp()).ui(
        editor='datetime'
    )

    def get_buttons(self):
        self.message.set(
            "<h2>Clear %s jobs</h2>" % self._site.name()
        )
        return ["Clear", "Cancel"]
    
    def run(self, button):
        queue = self._site.get_queue()
        status = self.status.get()
        type = self.type.get()

        if status == "All" and type == "All":
            queue.clear()
        else:
            for job in queue.jobs(status=self.status.get()):
                if job.date.get() < self.emitted_before.get():
                    queue.remove(job.name())

        queue.touch()


class RemoveEnvironmentVariableAction(flow.Action):

    ICON = ('icons.gui', 'remove-symbol')

    _variable = flow.Parent().ui(hidden=True)
    _collection = flow.Parent(2).ui(hidden=True)

    def get_buttons(self):
        return ['Confirm', 'Cancel']

    def run(self, button):
        if button == 'Cancel':
            return

        collection = self._collection
        self._collection.remove(self._variable.name())
        collection.touch()


class EnvironmentVariable(flow.Object):

    site = flow.Parent(2)
    
    variable = flow.Param("")
    value_windows = flow.Param('')
    value_linux   = flow.Param('')
    value_darwin  = flow.Param('')
    value = flow.Computed(cached=True)

    remove_variable = flow.Child(RemoveEnvironmentVariableAction).ui(hidden=True)

    def compute_child_value(self, child_value):
        if child_value is self.value:
            value = None
            # Get the operative system
            _os = platform.system()
            if _os == "Linux":
                value = self.value_linux.get()
            elif _os == "Windows":
                value = self.value_windows.get()
            elif _os == "Darwin":
                value = self.value_darwin.get()
            else:
                self.root().session().log_error('Unrecognized operative system')
                value = None

            self.value.set(value)


class AddEnvironmentVariableAction(flow.Action):

    _vars_collection = flow.Parent().ui(hidden=True)
    variable = flow.SessionParam("")
    value_windows = flow.SessionParam('')
    value_linux   = flow.SessionParam('')
    value_darwin  = flow.SessionParam('')

    def get_buttons(self):
        return ['Create', 'Cancel']
    
    def revert_params_to_defaults(self):
        self.variable.revert_to_default()
        self.value_windows.revert_to_default()
        self.value_linux.revert_to_default()
        self.value_darwin.revert_to_default()

    def run(self, button):
        var_name = self.variable.get().replace(' ', '_').upper()
        value_windows = self.value_windows.get()
        value_linux = self.value_linux.get()
        value_darwin = self.value_darwin.get()

        if button == 'Cancel':
            self.revert_params_to_defaults()
            return

        elif (len(var_name) == 0) or (len(value_windows) == 0 and len(value_linux) == 0 and len(value_darwin) == 0) or (self._vars_collection.has_mapped_name(var_name)):
            self.revert_params_to_defaults()
            return

        else:
            new_var = self._vars_collection.add(var_name)
            
            new_var.variable.set(var_name)
            new_var.value_windows.set(value_windows)
            new_var.value_linux.set(value_linux)
            new_var.value_darwin.set(value_darwin)
            self.revert_params_to_defaults()
            self._vars_collection.touch()


class SiteEnvironment(flow.Map):
    
    add_environment_variable = flow.Child(AddEnvironmentVariableAction)

    @classmethod
    def mapped_type(cls):
        return EnvironmentVariable


    def columns(self):
        return ['Variable', 'Value']

    def _fill_row_cells(self, row, item):
        row['Variable'] = item.variable.get()
        row['Value'] = item.value.get()

from .users import PresetValue

class SiteJobsPoolNames(PresetValue):

    DEFAULT_EDITOR = 'choice'
    
    def choices(self):
        site = self.root().project().get_current_site()
        return ['default'] + site.pool_names.get()


class WorkingSite(Site):
    
    _site_map = flow.Parent()
    
    site_type = flow.Param("Studio", StaticSiteTypeChoices)
    request_files_from_anywhere = flow.BoolParam(False).ui(
        tooltip=(
            "Allow the site to request files for any other site. "
            "Temporary option as long as synchronisation is manual."
        )
    )
    is_kitsu_admin = flow.BoolParam(False)
    auto_upload_kitsu_playblasts = flow.BoolParam(True)

    root_folder = flow.Computed(cached=True)
    root_windows_folder = flow.Param()
    root_linux_folder = flow.Param()
    root_darwin_folder = flow.Param()
    
    ordered_site_names = flow.Computed(cached=True)
    custom_site_order = flow.Param('').watched().ui(
        label='Custom order',
        tooltip='Manage order in which sites are listed in the interface')
    
    sync_dl_max_connections = flow.IntParam(1)
    sync_up_max_connections = flow.IntParam(1)
    pool_names = flow.OrderedStringSetParam()

    queue = flow.Child(Queue)

    configuration = flow.Child(ConfigureWorkingSite)
    goto_queue = flow.Child(GotoQueue).ui(
        label="Show job queue"
    )
    clear_queue = flow.Child(ClearQueueAction)
    site_environment = flow.Child(SiteEnvironment)

    with flow.group('AfterEffects settings'):
        ae_render_settings     = flow.Param().ui(label='Render settings')
        ae_output_module_movie = flow.Param().ui(label='Output module (movie)')
        ae_output_module_audio = flow.Param().ui(label='Output module (audio)')

    def compute_child_value(self, child_value):
        if child_value is self.root_folder:
            root_dir = None
            # Get the operative system
            _os = platform.system()
            if _os == "Linux":
                root_dir = self.root_linux_folder.get()
            elif _os == "Windows":
                root_dir = self.root_windows_folder.get()
            elif _os == "Darwin":
                root_dir = self.root_darwin_folder.get()
            else:
                print("ERROR: Unrecognized operative system ?")
            
            if not root_dir or not os.path.exists(root_dir):
                print("WARNING: ROOT_DIR path DOES NOT EXISTS")

            child_value.set(root_dir)

        elif child_value is self.configured:
            self.configured.set(bool(self.root_folder.get()))
        elif child_value is self.ordered_site_names:
            self.ordered_site_names.set(self._compute_ordered_site_names())
    
    def child_value_changed(self, child_value):
        if child_value is self.custom_site_order:
            self.ordered_site_names.touch()
    
    def _compute_ordered_site_names(self):
        names_as_string = self.custom_site_order.get()
        names = names_as_string.replace(' ','').split(',')
        
        if names and names[0] == '':
            names = []
        
        unordered_names = self._site_map.mapped_names()

        for n in names:
            unordered_names.remove(n)
        
        return names + unordered_names
    
    def get_queue(self):
        return self.queue.job_list
    
    def get_jobs(self, type=None, status=None):
        return self.get_queue().jobs(type=type, status=status)
    
    def count_jobs(self, type=None, status=None):
        return self.get_queue().count(type=type, status=status)


class ClearSiteQueues(flow.Action):

    ICON = ('icons.libreflow', 'clean')

    _sites = flow.Parent()
    emitted_since = flow.Param(0.0)

    def get_buttons(self):
        return ['Clear', 'Cancel']
    
    def run(self, button):
        if button == 'Cancel':
            return
        
        for s in self._sites.mapped_items():
            for j in s.get_jobs(status='PROCESSED'):
                c = s.queue.job_list.get_entity_store().get_collection(
                    s.queue.job_list.collection_name()
                )
                c.delete_many({
                    'date': {'$lt': time.time() - self.emitted_since.get()},
                    'status': 'PROCESSED'
                })

        self.root().project().synchronization.touch()


class WorkingSites(SiteMap):

    ICON = ('icons.gui', 'home')

    clear_site_queues = flow.Child(ClearSiteQueues)

    @classmethod
    def mapped_type(cls):
        return flow.injection.injectable(WorkingSite)

    def _configure_child(self, child):
        if child.name() == "default":
            child.short_name.set("dft")
            child.description.set("Default working site")
            child.site_type.set("Studio")
        else:
            super(WorkingSites, self)._configure_child(child)

    def _fill_row_style(self, style, item, row):
        super(WorkingSites, self)._fill_row_style(style, item, row)

        if item.site_type.get() == "User":
            style['icon'] = ('icons.gui', 'user')
        else:
            style['icon'] = ('icons.gui', 'home')


class ConfigureExchangeSite(ConfigureSite):

    server_url = flow.SessionParam("http://")
    server_login = flow.SessionParam("")
    server_password = flow.SessionParam("").ui(editor='password')
    bucket_name = flow.SessionParam('')

    def _fill_action_fields(self):
        super(ConfigureExchangeSite, self)._fill_action_fields()
        self.server_url.set(self._site.server_url.get())
        self.server_login.set(self._site.server_login.get())
        self.server_password.set(self._site.server_password.get())
        self.bucket_name.set(self._site.bucket_name.get())

    def _configure_site(self, site):
        super(ConfigureExchangeSite, self)._configure_site(site)
        site.server_url.set(self.server_url.get())
        site.server_login.set(self.server_login.get())
        site.server_password.set(self.server_password.get())
        site.bucket_name.set(self.bucket_name.get())


class TestConnectionAction(flow.Action):

    _exchange = flow.Parent()

    def needs_dialog(self):
        return True

    def get_buttons(self):
        self.message.set('Clik the button to test the connection.')
        return ['Test']

    def run(self, button):
        ret = self._exchange.sync_manager.check_connection()

        if ret is not None:
            self.message.set(
                '<font color=red>Connection error:</font><br>'
                f'<pre>{ret}</pre>'
            )
        else:
            self.message.set(
                'Connection looks <b>OK</b>'
            )
        return self.get_result(close=False)


class ExchangeSite(Site):

    ICON = ('icons.libreflow', 'exchange')

    server_url = flow.Param("")
    server_login = flow.Param("")
    server_password = flow.Param("").ui(editor='password')
    bucket_name = flow.Param('')
    enable_tls = flow.BoolParam(False)

    sync_manager = flow.Child(SyncManager).ui(hidden=True)

    configuration = flow.Child(ConfigureExchangeSite)
    test_connection = flow.Child(TestConnectionAction)

    def compute_child_value(self, child_value):
        if child_value is self.configured:
            self.configured.set(
                (
                    self.server_url.get()
                    and self.server_login.get()
                    and self.server_password.get()
                    and self.bucket_name.get()
                )
            )


class ExchangeSites(SiteMap):
    
    ICON = ('icons.libreflow', 'exchange')

    @classmethod
    def mapped_type(cls):
        return ExchangeSite

    def mapped_names(self, page_num=0, page_size=None):
        return ["default_exchange"] + super(SiteMap, self).mapped_names(page_num, page_size)

    def _configure_child(self, child):
        if child.name() == "default_exchange":
            child.short_name.set("dftx")
            child.description.set("Default exchange site")
        else:
            super(ExchangeSites, self)._configure_child(child)

    def _get_mapped_item_type(self, mapped_name):
        if mapped_name == "default_exchange":
            return self.mapped_type()

        return super(SiteMap, self)._get_mapped_item_type(mapped_name)


class SyncSiteStatus(flow.Object):
    status = flow.Param("NotAvailable", StaticSiteSyncStatusChoices)


class SyncMap(flow.DynamicMap):
    version = flow.Parent()

    @classmethod
    def mapped_type(cls):
        return SyncSiteStatus

    def mapped_names(self, page_num=0, page_size=None):
        sites_data = self.root().project().admin.multisites.sites_data.get()
        return sites_data.keys()

    def columns(self):
        return ['Name', 'Status']

    def _fill_row_cells(self, row, item):
        row['Status'] = item.status.get()
        name = item.name()

        if name == self.version.site.get():
            name += " (S)"
        
        row['Name'] = name


class RequestedRevisions(flow.DynamicMap):
    
    STYLE_BY_STATUS = {
        'Available': ('#45cc3d', ('icons.libreflow', 'blank')),
        'Requested': ('#b9c2c8', ('icons.libreflow', 'exclamation-sign-colored')),
        'NotAvailable': ('#cc3b3c', ('icons.libreflow', 'blank')),
        'Undefined': ('#cc3b3c', ('icons.libreflow', 'unavailable')),
    }
    
    request_action = flow.Parent()
    oids = flow.SessionParam([]).ui(editor='set')
    
    def mapped_names(self, page_num=0, page_size=None):
        return self.oids.get()
    
    def get_mapped(self, name):
        if not self.has_mapped_name(name):
            raise MappedNameError(self.oid(), name)

        try:
            obj = self._mng.get_object(name)
        except ValueError:
            raise

        return obj
    
    def columns(self):
        return ['Revision']
    
    def _fill_row_cells(self, row, item):
        row['Revision'] = item.oid()
    
    def _fill_row_style(self, style, item, row):
        status = 'Undefined'
        
        if hasattr(item, 'request_as'):
            status = item.get_sync_status(site_name=self.request_action.sites.target_site.get())
            
        style['Revision_foreground-color'] = self.STYLE_BY_STATUS[status][0]
        style['Revision_icon'] = self.STYLE_BY_STATUS[status][1]


class RequestRevisions(flow.Action):
    
    ICON = ('icons.gui', 'share-option')

    pattern = flow.SessionParam("").ui(
        placeholder="Revision oid pattern"
    )
    
    sites = flow.Child(SiteSelection).ui(expanded=True)
    revisions = flow.Child(RequestedRevisions).ui(expanded=True)
    
    def allow_context(self, context):
        current_site = self.root().project().get_current_site()
        return context and current_site.request_files_from_anywhere.get()
    
    def _get_oid(self, pattern_str):
        oids = []
        patterns = []
        
        for pattern in pattern_str.split(';'):
            patterns += self.resolve_pattern(pattern)
        
        for pattern in patterns:
            oids += self.glob(self.root().project().oid(), pattern, 0)
        
        return oids

    def get_buttons(self):
        self.sites.source_site.set(self.root().project().get_current_site().name())
        target_site_choices = self.sites.target_site.choices()
        
        if not target_site_choices:
            msg = self.message.get()
            if msg is None:
                msg = ''
            
            msg += (
                '<h3><font color="#D5000D">Making requests is not possible since '
                'there is no other site defined for this project.</font></h3>'
            )
            self.message.set(msg)
            
            return ['Close']
        
        self.sites.target_site.set(target_site_choices[0])

        return ["Request", "Refresh revision list", "Close"]
    
    def ls(self, root_oid):
        related_info, mapped_names = self.root().session().cmds.Flow.ls(root_oid)
        relation_oids = [rel_info[0] for rel_info in related_info]
        mapped_oids = ["%s/%s" % (root_oid, name) for name in mapped_names]
        
        return relation_oids + mapped_oids
    
    def get_last_publication(self, file_oid):
        o = self.root().get_object(file_oid)
        
        try:
            head = o.get_head_revision()
        except AttributeError:
            return None
        
        if not head:
            return None
        
        return head.name()
    
    def glob(self, root_oid, pattern, level):
        if level >= pattern.count("/") - 1:
            return [root_oid]

        matches = []
        level_pattern = "/".join(pattern.split("/")[:level + 3])

        if level_pattern.endswith("[last]"):
            file_oid = self.root().session().cmds.Flow.resolve_path(root_oid + "/../..")
            head_name = self.get_last_publication(file_oid)

            if not head_name:
                return []
            
            pattern = pattern.replace("[last]", head_name)
            level_pattern = level_pattern.replace("[last]", head_name)
        
        for oid in self.ls(root_oid):
            if fnmatch.fnmatch(oid, level_pattern):
                matches += self.glob(oid, pattern, level + 1)
        
        return matches
    
    def resolve_pattern(self, pattern_oid):
        '''
        Returns a list of patterns which correspond to all substitution combinations of `pattern_oid`.
        Substitute string are specified in `pattern_oid` between brackets and separated with comas (`{s0, s1, ...}`).
        '''
        match = re.search(r'{[^{}]+}', pattern_oid)

        if not match:
            return [pattern_oid]

        substitutes = match.group()[1:-1].split(',')
        substitutes = [sub.replace(' ', '') for sub in substitutes]

        sub_oids = [pattern_oid.replace(match.group(), sub, 1) for sub in substitutes]
        oids = []

        for sub_oid in sub_oids:
            oids += self.resolve_pattern(sub_oid)
        
        return oids
    
    def run(self, button):
        if button == "Close":
            return
        elif button == "Refresh revision list":
            self.revisions.oids.revert_to_default()
            oids = self._get_oid(self.pattern.get())
            self.revisions.oids.set(oids)
            self.revisions.touch()
            
            return self.get_result(close=False)
        
        source_site = self.sites.source_site.get()
        target_site = self.sites.target_site.get()
        
        for obj in self.revisions.mapped_items():
            # Skip objects which are not file revisions
            if not hasattr(obj, 'request_as'):
                continue
            # Skip revisions which are not available on selected source site
            if obj.get_sync_status(site_name=source_site) != 'Available':
                continue

            # Skip revisions which are already available on selected target site
            status = obj.get_sync_status(site_name=target_site)
            if status == 'Requested' or status == 'Available':
                continue
            
            obj.request_as.sites.source_site.set(source_site)
            obj.request_as.sites.target_site.set(target_site)
            obj.request_as.include_dependencies.set(False)
            obj.request_as.run(None)
        
        self.revisions.touch()

        return self.get_result(close=False)


class SiteMultiSelection(flow.Object):
    
    source_site = flow.Param('Auto select', ActiveSiteAutoSelectChoiceValue).watched()
    target_sites = flow.Param([], ActiveSitesMultichoiceValue)
    
    def child_value_changed(self, child_value):
        if child_value is self.source_site:
            source_site = self.source_site.get()
            
            if source_site == 'Auto select':
                self.target_sites.exclude_choice.set(None)
            else:
                self.target_sites.exclude_choice.set(self.source_site.get())
            
            self.target_sites.touch()


class MultiRequestedRevisions(RequestedRevisions):
    
    STYLE_BY_STATUS = {
        'Available': ('#45cc3d', ('icons.libreflow', 'checked-symbol-colored')),
        'Requested': ('#b9c2c8', ('icons.libreflow', 'exclamation-sign-colored')),
        'NotAvailable': ('#cc3b3c', ('icons.libreflow', 'blank')),
        'Undefined': ('#cc3b3c', ('icons.libreflow', 'unavailable')),
    }
    
    request_action = flow.Parent()
    source_sites = flow.SessionParam([]).ui(hidden=True)
    
    def columns(self):
        site_names = self.request_action.sites.target_sites.get() + self.source_sites.get()
        site_names = list(dict.fromkeys(site_names))

        return ['Revision'] + site_names
    
    def _fill_row_cells(self, row, item):
        row['Revision'] = item.oid()
        
        for col in self.columns()[1:]:
            row[col] = ''
    
    def _fill_row_style(self, style, item, row):
        style['Revision_icon'] = ('icons.libreflow', 'blank')
        
        if not hasattr(item, 'request_as'):
            style['Revision_icon'] = self.STYLE_BY_STATUS['Undefined'][1]
            return
        
        site_names = self.request_action.sites.target_sites.get() + self.source_sites.get()
        site_names = list(dict.fromkeys(site_names))

        for site_name in site_names:
            status = item.get_sync_status(site_name=site_name)
            style[site_name + '_icon'] = self.STYLE_BY_STATUS[status][1]


class MultiRequestRevisions(RequestRevisions):
    
    ICON = ('icons.libreflow', 'multi-share-option')
    
    sites = flow.Child(SiteMultiSelection).ui(expanded=True)
    revisions = flow.Child(MultiRequestedRevisions).ui(expanded=True)
    
    def get_buttons(self):
        self.sites.source_site.revert_to_default()
        target_sites_choices = self.sites.target_sites.choices()
        
        if not target_sites_choices:
            msg = self.message.get()
            if msg is None:
                msg = ''
            
            msg += (
                '<h3><font color="#D5000D">Making requests is not possible since '
                'there is no other site defined for this project.</font></h3>'
            )
            self.message.set(msg)
            
            return ['Close']

        return ['Request', 'Refresh revision list', 'Close']
    
    def run(self, button):
        if button == 'Close':
            return
        elif button == 'Refresh revision list':
            self.revisions.oids.revert_to_default()
            oids = self._get_oid(self.pattern.get())
            self.revisions.oids.set(oids)

            source_sites = []
            for oid in oids:
                site_name = self.root().session().cmds.Flow.get_value(oid + '/site')
                source_sites.append(site_name)
            
            self.revisions.source_sites.set(source_sites)
            self.revisions.touch()
            
            return self.get_result(close=False)
        
        source_site = self.sites.source_site.get()
        auto_select_enabled = (source_site == 'Auto select')
        
        for obj in self.revisions.mapped_items():
            # Skip objects which are not file revisions
            if not hasattr(obj, 'request_as'):
                continue
            # Skip revisions which are not available on selected source site
            if not auto_select_enabled and obj.get_sync_status(site_name=source_site) != 'Available':
                continue
            
            for target_site in self.sites.target_sites.get():
                # Skip revisions which are already available on selected target site
                status = obj.get_sync_status(site_name=target_site)
                if status == 'Requested' or status == 'Available':
                    continue
                
                if not auto_select_enabled:
                    obj.request_as.sites.source_site.set(source_site)
                
                obj.request_as.sites.target_site.set(target_site)
                obj.request_as.sites.auto_select_enabled.set(auto_select_enabled)
                obj.request_as.include_dependencies.set(False)
                obj.request_as.run(None)
        
        self.revisions.touch()

        return self.get_result(close=False)


class UploadRevision(flow.Action):

    ICON = ('icons.libreflow', 'upload')
    
    _revision = flow.Parent()
    _revisions = flow.Parent(2)
    
    def needs_dialog(self):
        return self._revision.get_sync_status(exchange=True) == 'Available'
    
    def allow_context(self, context):
        return (
            context
            and not self._revision.is_working_copy()
            and self._revision.get_sync_status() == 'Available'
        )
    
    def get_buttons(self):  
        self.message.set((
            '<h3>Revision already on the exchange server</h3>'
            'Upload and overwrite it on the server anyway ?'
        ))
        return ['Confirm', 'Cancel']
    
    def run(self, button):
        if button == 'Cancel':
            return

        current_site = self.root().project().get_current_site()
        # Add an upload job for the current site
        job = current_site.get_queue().submit_job(
            job_type='Upload',
            init_status='WAITING',
            emitter_oid=self._revision.oid(),
            user=self.root().project().get_user_name(),
            studio=current_site.name(),
        )
        sync_manager = self.root().project().get_sync_manager()
        sync_manager.process(job)
        
        self._revisions.touch()


class DownloadRevision(flow.Action):

    ICON = ('icons.libreflow', 'download')
    
    _revision = flow.Parent()
    _revisions = flow.Parent(2)

    def needs_dialog(self):
        return self._revision.get_sync_status() == 'Available'
    
    def get_buttons(self):  
        self.message.set((
            '<h3>Revision already available</h3>'
            'Download and overwrite it locally anyway ?'
        ))
        return ['Confirm', 'Cancel']
    
    def allow_context(self, context):
        return (
            context
            and not self._revision.is_working_copy()
            and self._revision.get_sync_status(exchange=True) == 'Available'
        )
    
    def run(self, button):
        if button == 'Cancel':
            return

        current_site = self.root().project().get_current_site()
        # Add an upload job for the current site
        job = current_site.get_queue().submit_job(
            job_type='Download',
            init_status='WAITING',
            emitter_oid=self._revision.oid(),
            user=self.root().project().get_user_name(),
            studio=current_site.name(),
        )
        self._revision.set_sync_status('Requested', current_site.name())
        sync_manager = self.root().project().get_sync_manager()
        sync_manager.process(job)
        
        self._revisions.touch()



import os

from kabaret import flow
from kabaret.subprocess_manager.flow import RunAction


# RunActions
# -----------------


class OpenFileAction(RunAction):

    _file = flow.Parent()

    def needs_dialog(self):
        return False

    @classmethod
    def supported_extensions(cls):
        """
        Supported file extensions.

        Return None by default to allow any extension.
        """
        return None

    def allow_context(self, context):
        ext = os.path.splitext(self._file.path.get())[1]
        valid_ext = (
            self.supported_extensions() is None or ext in self.supported_extensions()
        )

        return context and context.endswith(".map") and valid_ext

    def extra_argv(self):
        return [self._file.path.get()]


class OpenWithBlenderAction(OpenFileAction):
    @classmethod
    def supported_extensions(cls):
        return [".blend"]

    def runner_name_and_tags(self):
        return "Blender", []


class OpenWithKritaAction(OpenFileAction):
    @classmethod
    def supported_extensions(cls):
        return [".kra", ".png", ".jpg"]

    def runner_name_and_tags(self):
        return "Krita", []



from kabaret.app.ui.gui.widgets.flow.flow_view import (
    CustomPageWidget,
    QtWidgets,
    QtCore,
    QtGui,
)
from fnmatch import fnmatch
from collections import defaultdict


class TaskFileItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, task_item, tree, file):
        super(TaskFileItem, self).__init__(task_item)
        self._tree = tree
        self._task_item = task_item
        self.file = None
        
        self.combo_box_revisions = QtWidgets.QComboBox()
        self._tree.setItemWidget(self, 1, self.combo_box_revisions)
        
        self.set_file(file)
    
    def set_file(self, file):
        self.file = file
        self._update()
    
    def on_checkstate_toggled(self, column):
        pass
    
    def submit_blender_playblast_job(self):
        f = self.file
        kwargs = dict(
            shot_name=f['shot'],
            dept_name=f['department'],
            file_name=f['name'],
            revision_name=self.combo_box_revisions.currentText(),
        )
        extension = f['name'].split('.')[1]
        
        if extension == 'blend':
            kwargs['use_simplify'] = (self.checkState(2) == QtCore.Qt.Checked)
            kwargs['reduce_textures'] = (self.checkState(3) == QtCore.Qt.Checked)
        
        return self._tree.parentWidget().submit_blender_playblast_job(extension, **kwargs)
    
    def _update(self):
        f = self.file
        self.setText(0, f['name'])
        
        names, statuses = zip(*f['revisions'])
        self.combo_box_revisions.addItems(names)
        self.combo_box_revisions.setCurrentText(f['default_revision'])
        
        for i, status in enumerate(statuses):
            if status != 'Available':
                self.combo_box_revisions.setItemData(i, False, QtGui.Qt.UserRole-1)


class ShotTaskItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, shot_item, tree, task):
        super(ShotTaskItem, self).__init__(shot_item)
        self._shot_item = shot_item
        self._tree = tree
        self.task = None
        
        self.set_task(task)
    
    def set_task(self, task):
        self.task = task
        self._update()
    
    def on_checkstate_toggled(self, column):
        for i in range(self.childCount()):
            self.child(i).setCheckState(column, self.checkState(column))
    
    def _update(self):
        d = self.task
        self.setText(0, d['name'])
        
        file_item_type = self._tree.file_item_type()
        
        # Loop over task files
        for file in d['files']:
            file_item_type(self, self._tree, file)


class ShotItem(QtWidgets.QTreeWidgetItem):

    def __init__(self, shots_item, tree, shot):
        super(ShotItem, self).__init__(shots_item)
        self._tree = tree
        self.shot = None
        
        self.set_shot(shot)
    
    def set_shot(self, shot):
        self.shot = shot
        self._update()
    
    def on_checkstate_toggled(self, column):
        # Uncheck all tasks if shot is unchecked
        if self.checkState(column) == QtCore.Qt.Unchecked:
            for i in range(self.childCount()):
                self.child(i).setCheckState(column, QtCore.Qt.Unchecked)
            
            return
        
        # Otherwise, check tasks according to global tasks checkstates
        for i in range(self.childCount()):
            check_state = self._tree.get_task_item_checkstate(column, i)
            self.child(i).setCheckState(column, check_state)
    
    def _update(self):
        s = self.shot
        self.setText(0, s['name'])
        
        for i in range(self._tree.header().count()):
            self.setBackgroundColor(i, QtGui.QColor(60, 60, 60))
        
        # Loop over shot tasks
        for task in s['tasks']:
            task_item = ShotTaskItem(self, self._tree, task)


class TaskItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, tasks_item, tree, task_name):
        super(TaskItem, self).__init__(tasks_item)
        self._tree = tree
        self.tasks_item = tasks_item
        
        self.setText(0, task_name)
        
        for i in range(self._tree.header().count()):
            self.setBackgroundColor(i, QtGui.QColor(60, 60, 60))
    
    def on_checkstate_toggled(self, column):
        index = self.tasks_item.indexOfChild(self)
        self._tree.set_task_items_checkstate(self.checkState(column), column, index)


class TasksItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, tree):
        super(TasksItem, self).__init__(tree)
        self._tree = tree
        
        self._update()
    
    def on_checkstate_toggled(self, column):
        for i in range(self.childCount()):
            self.child(i).setCheckState(column, self.checkState(column))
    
    def _update(self):
        self.setText(0, 'Tasks')
        
        for i in range(self._tree.header().count()):
            self.setBackgroundColor(i, QtGui.QColor(76, 80, 82))
            font = self.font(i)
            font.setWeight(QtGui.QFont.DemiBold)
            self.setFont(i, font)
        
        task_names = self._tree.parentWidget().get_shot_task_names()
        for task_name in task_names:
            task_item = TaskItem(self, self._tree, task_name)
        
        self.setExpanded(True)


class ShotsItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, tree):
        super(ShotsItem, self).__init__(tree)
        self._tree = tree
        
        self._update()
    
    def on_checkstate_toggled(self, column):
        for i in range(self.childCount()):
            self.child(i).setCheckState(column, self.checkState(column))
    
    def _update(self):
        self.setText(0, 'Shots')
        
        for i in range(self._tree.header().count()):
            self.setBackgroundColor(i, QtGui.QColor(76, 80, 82))
            font = self.font(i)
            font.setWeight(QtGui.QFont.DemiBold)
            self.setFont(i, font)
        
        shot_files_data = self._tree.parentWidget().get_shots_data()
        for shot in shot_files_data:
            shot_item = ShotItem(self, self._tree, shot)
        
        self.setExpanded(True)


class ShotSelector(QtWidgets.QTreeWidget):
    
    @classmethod
    def file_item_type(cls):
        return TaskFileItem
    
    def __init__(self, parent, session):
        super(ShotSelector, self).__init__(parent)
        self.session = session

        self.setHeaderLabels(self.get_columns())
        self.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Stretch)
        self.header().setStretchLastSection(False)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        self.setFixedHeight(350)
        
        self.tasks = None
        self.shots = None
        self.task_name_to_index = {}
        self.refresh()
        
        # self.setMinimumHeight(self.sizeHintForRow(0) * (self.shots.childCount() + self.tasks.childCount()))
        
        self.itemChanged.connect(self.on_item_changed)
    
    def get_columns(self):
        return ('Name', 'Revision', 'Use simplify', 'Reduce textures')
    
    def on_item_changed(self, item, column):
        item.on_checkstate_toggled(column)
    
    def set_task_items_checkstate(self, state, column, index):
        for i in range(self.shots.childCount()):
            shot_item = self.shots.child(i)
            
            if shot_item.checkState(0) == QtCore.Qt.Checked:
                task_item = shot_item.child(index)
                task_item.setCheckState(column, state)
    
    def get_task_item_checkstate(self, column, index):
        return self.tasks.child(index).checkState(column)
    
    def _do_refresh(self, item):
        item.setCheckState(0, QtCore.Qt.Unchecked)
        item.setCheckState(2, QtCore.Qt.Unchecked)
        item.setCheckState(3, QtCore.Qt.Unchecked)
    
    def refresh_child(self, item):
        self._do_refresh(item)
        
        if item.childCount() == 0:
            return
        
        for i in range(item.childCount()):
            self.refresh_child(item.child(i))
    
    def refresh(self):
        self.clear()
        
        self.tasks = TasksItem(self)
        self.shots = ShotsItem(self)
        
        for i in range(self.topLevelItemCount()):
            self.refresh_child(self.topLevelItem(i))


class ProcessSequenceFilesWidget(CustomPageWidget):

    def get_shots_data(self):
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name='ensure_files_data',
            args={}, kwargs={}
        )
    
    def get_shot_task_names(self):
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name='get_shot_task_names',
            args={}, kwargs={}
        )

    def _close_view(self):
        view = self.parentWidget().page.view
        view.close()


class PlayblastItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, task_item, tree, file):
        super(PlayblastItem, self).__init__(task_item)
        self._tree = tree
        self._task_item = task_item
        self.file = None
        
        self.combo_box_revision = QtWidgets.QComboBox()
        self.combo_box_target_status = QtWidgets.QComboBox()
        self.combo_box_task_type = QtWidgets.QComboBox()
        self.lineedit_comment = QtWidgets.QLineEdit()
        self.lineedit_comment.setPlaceholderText('...')
        
        self._tree.setItemWidget(self, 1, self.combo_box_revision)
        self._tree.setItemWidget(self, 2, self.combo_box_task_type)
        self._tree.setItemWidget(self, 4, self.combo_box_target_status)
        self._tree.setItemWidget(self, 5, self.lineedit_comment)
        
        self.set_file(file)
    
    def set_file(self, file):
        self.file = file
        self._update()
    
    def on_checkstate_toggled(self, column):
        pass
    
    def update_target_status(self, target_status):
        self.combo_box_target_status.setCurrentText(target_status)
    
    def _update(self):
        f = self.file
        self.setText(0, f['name'])
        self.setText(3, f['current_task_status'])
        
        names, revision_statutes = zip(*f['revisions'])
        self.combo_box_revision.addItems(names)
        self.combo_box_revision.setCurrentText(f['default_revision'])
        self.combo_box_task_type.addItems(f['task_types'])
        self.combo_box_task_type.setCurrentText(f['default_task_type'])
        task_statutes = self._tree.parentWidget().get_task_statutes()
        self.combo_box_target_status.addItems(task_statutes)
        # self.combo_box_target_status.setCurrentText(f['default_task_status'])
        self.combo_box_target_status.setCurrentText(task_statutes[0])
        
        for i, status in enumerate(revision_statutes):
            if status != 'Available':
                self.combo_box_revision.setItemData(i, False, QtGui.Qt.UserRole-1)
    
    def upload_to_kitsu(self):
        comment = self.lineedit_comment.text()
        task_name = self.parent().text(0)
        task_comment = self._tree.get_task_comment(task_name)
        
        if task_comment:
            comment = task_comment + '\n\n' + comment
        
        self._tree.parentWidget().upload_playblast_to_kitsu(
            file_oid=self.file['oid'],
            revision_name=self.combo_box_revision.currentText(),
            task_type=self.combo_box_task_type.currentText(),
            target_task_status=self.combo_box_target_status.currentText(),
            comment=comment
        )


class UploadPlayblastTaskItem(TaskItem):
    
    def __init__(self, tasks_item, tree, task_name, kitsu_task_type):
        super(UploadPlayblastTaskItem, self).__init__(tasks_item, tree, task_name)
        
        self.combo_box_target_status = QtWidgets.QComboBox()
        self.lineedit_comment = QtWidgets.QLineEdit()
        self.lineedit_comment.setPlaceholderText('...')
        
        self._tree.setItemWidget(self, 4, self.combo_box_target_status)
        self._tree.setItemWidget(self, 5, self.lineedit_comment)
        
        self.setText(2, kitsu_task_type)
        self._update()
        
        self.combo_box_target_status.currentTextChanged.connect(self.on_target_status_changed)
    
    def get_comment(self):
        return self.lineedit_comment.text()
    
    def on_target_status_changed(self, target_status):
        self._tree.update_target_statutes(self.text(0), target_status)
    
    def _update(self):
        task_statutes = self._tree.parentWidget().get_task_statutes()
        self.combo_box_target_status.addItems(task_statutes)
        self.combo_box_target_status.setCurrentText(task_statutes[0])


class UploadPlayblastTasksItem(TasksItem):
    
    def _update(self):
        self.setText(0, 'Tasks')
        
        for i in range(self._tree.header().count()):
            self.setBackgroundColor(i, QtGui.QColor(76, 80, 82))
            font = self.font(i)
            font.setWeight(QtGui.QFont.DemiBold)
            self.setFont(i, font)
        
        task_names = self._tree.parentWidget().get_shot_task_names()
        for task_name in task_names:
            kitsu_task_type = self._tree.parentWidget().get_kitsu_task_type(task_name)
            task_item = UploadPlayblastTaskItem(self, self._tree, task_name, kitsu_task_type)
        
        self.setExpanded(True)


class UploadPlayblastSelector(ShotSelector):
    
    @classmethod
    def file_item_type(cls):
        return PlayblastItem
    
    def get_columns(self):
        return ('Shot', 'Revision', 'Kitsu task', 'Current status', 'Target status', 'Comment')
    
    def get_task_comment(self, task_name):
        task_index = self.task_name_to_index[task_name]
        return self.tasks.child(task_index).get_comment()
    
    def update_target_statutes(self, task_name, target_status):
        task_index = self.task_name_to_index[task_name]
        
        for i in range(self.shots.childCount()):
            shot_item = self.shots.child(i)
            task_item = shot_item.child(task_index)
            
            for j in range(task_item.childCount()):
                file_item = task_item.child(j)
                file_item.update_target_status(target_status)
    
    def _do_refresh(self, item):
        item.setCheckState(0, QtCore.Qt.Unchecked)
    
    def refresh(self):
        self.clear()
        self.task_name_to_index.clear()
        
        self.tasks = UploadPlayblastTasksItem(self)
        self.shots = ShotsItem(self)
        
        for i in range(self.topLevelItemCount()):
            self.refresh_child(self.topLevelItem(i))
        
        for i in range(self.tasks.childCount()):
            task_name = self.tasks.child(i).text(0)
            self.task_name_to_index[task_name] = i
        
        self.header().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.header().setStretchLastSection(True)


class UploadSequencePlayblastsWidget(ProcessSequenceFilesWidget):
    
    def build(self):
        self.playblast_selector = UploadPlayblastSelector(self, self.session)
        self.button_upload = QtWidgets.QPushButton('Upload')
        self.button_cancel = QtWidgets.QPushButton('Cancel')
        
        glo = QtWidgets.QGridLayout()
        glo.addWidget(self.playblast_selector, 0, 0, 3, 1)
        glo.addWidget(self.button_upload, 0, 1)
        glo.addWidget(self.button_cancel, 1, 1)
        glo.setSpacing(0)
        self.setLayout(glo)

        self.button_upload.clicked.connect(self.on_upload_button_clicked)
        self.button_cancel.clicked.connect(self.on_cancel_button_clicked)
    
    def get_task_statutes(self):
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name='get_task_statutes',
            args={}, kwargs={}
        )
    
    def get_kitsu_task_type(self, task_name):
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name='get_kitsu_task_type',
            args=[task_name], kwargs={}
        )
    
    def upload_playblast_to_kitsu(self, file_oid, revision_name, task_type, target_task_status, comment):
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name='upload_playblast_to_kitsu',
            args=[file_oid, revision_name, task_type, target_task_status, comment],
            kwargs={}
        )

    def on_upload_button_clicked(self):
        shots = self.playblast_selector.shots
        
        for i in range(shots.childCount()):
            shot_item = shots.child(i)
            
            for j in range(shot_item.childCount()):
                task_item = shot_item.child(j)
                
                for k in range(task_item.childCount()):
                    file_item = task_item.child(k)
                    
                    if file_item.checkState(0) == QtCore.Qt.Checked:
                        file_item.upload_to_kitsu()
        
        self._close_view()
    
    def on_cancel_button_clicked(self):
        self._close_view()


class RenderSequencePlayblastsWidget(ProcessSequenceFilesWidget):
    
    def build(self):
        self.shot_selector = ShotSelector(self, self.session)
        self.button_submit = QtWidgets.QPushButton('Submit playblasts')
        self.checkbox_select_all = QtWidgets.QCheckBox('Select all')
        self.combobox_pool = QtWidgets.QComboBox()
        self.lineedit_priority = QtWidgets.QLineEdit()
        
        self.button_submit.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.button_submit.setMaximumWidth(150)
        font = self.button_submit.font()
        font.setPointSize(11)
        font.setWeight(QtGui.QFont.DemiBold)
        self.button_submit.setFont(font)
        
        validator = QtGui.QIntValidator(1, 1000)
        self.lineedit_priority.setValidator(validator)
        self.lineedit_priority.setText('10')
        self.lineedit_priority.setFixedWidth(50)
        self.lineedit_priority.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        
        self.combobox_pool.addItems(self.get_job_pool_names())
        self.combobox_pool.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.combobox_pool.setFixedWidth(50)
        
        hlo = QtWidgets.QHBoxLayout()
        hlo.addWidget(self.shot_selector, 0, QtCore.Qt.AlignTop)
        
        vlo = QtWidgets.QVBoxLayout()
        label_settings = QtWidgets.QLabel('Global settings')
        label_settings.setFont(font)
        vlo.addWidget(label_settings, 0)
        vlo.addWidget(QtWidgets.QLabel('Priority'), 1)
        vlo.addWidget(self.lineedit_priority, 2)
        vlo.addWidget(QtWidgets.QLabel('Job pool'), 3)
        vlo.addWidget(self.combobox_pool, 4)
        vlo.addWidget(self.button_submit, 5)
        
        for i in range(5):
            vlo.setStretch(i, 0)
        vlo.insertStretch(5, 10)
        
        hlo.addLayout(vlo, 1)
        
        self.setLayout(hlo)

        self.button_submit.clicked.connect(self.on_submit_button_clicked)
    
    def get_job_pool_names(self):
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name='get_job_pool_names',
            args={}, kwargs={}
        )
    
    def submit_blender_playblast_job(self, file_extension, **kwargs):
        kwargs['pool_name'] = self.combobox_pool.currentText()
        kwargs['priority'] = int(self.lineedit_priority.text())
        
        if file_extension == 'blend':
            method_name = 'submit_blender_playblast_job'
        elif file_extension == 'aep':
            method_name = 'submit_afterfx_playblast_job'
        else:
            pass
        
        return self.session.cmds.Flow.call(
            oid=self.oid,
            method_name=method_name,
            args={}, kwargs=kwargs
        )
    
    def on_submit_button_clicked(self):
        shots = self.shot_selector.shots
        
        for i in range(shots.childCount()):
            shot_item = shots.child(i)
            
            for j in range(shot_item.childCount()):
                task_item = shot_item.child(j)
                
                for k in range(task_item.childCount()):
                    file_item = task_item.child(k)
                    
                    if file_item.checkState(0) == QtCore.Qt.Checked:
                        file_item.submit_blender_playblast_job()
        
        self._close_view()
    



import os
import re
import timeago
from datetime import datetime
from kabaret.app import resources
from kabaret.app.ui.gui.widgets.widget_view import DockedView, QtWidgets, QtGui, QtCore


def natural_sort_key(s, _nsre=re.compile('([0-9]+)')):
    '''Sort a string in a natural way
    https://stackoverflow.com/a/16090640'''
    return [int(text) if text.isdigit() else text.lower()
            for text in _nsre.split(s)]


def timeago_format(time):
    dt = datetime.fromtimestamp(time)
    dt = timeago.format(dt, datetime.now())
    return dt


def custom_format(time, format):
    dt = datetime.fromtimestamp(time)
    dt = dt.astimezone().strftime(format)
    return dt


class OutputHeader(QtWidgets.QWidget):
    
    def __init__(self, view, parent=None):
        super(OutputHeader, self).__init__(parent)
        self.view = view
        
        self.setFixedHeight(80)
        
        # Define widgets to display current runner data
        self._label_index = QtWidgets.QLabel()
        self._label_name = QtWidgets.QLabel()
        self._label_description = QtWidgets.QLabel()
        self._label_update_time = QtWidgets.QLabel()
        self._label_pid = QtWidgets.QLabel()
        self._textedit_cmd = QtWidgets.QPlainTextEdit()
        self._button_copy_cmd = QtWidgets.QPushButton(
            resources.get_icon(('icons.libreflow', 'clipboard')), ''
        )
        
        # Labels
        font = self._label_index.font()
        font.setPointSize(15)
        font.setWeight(QtGui.QFont.DemiBold)
        self._label_description.setFont(font)
        self._label_name.setFont(font)
        font.setPointSize(8)
        font.setWeight(QtGui.QFont.Normal)
        self._label_update_time.setFont(font)
        self._label_update_time.setStyleSheet("QLabel { color : #aaa; }")
        self._label_index.setFont(font)
        self._label_index.setStyleSheet("QLabel { color : #aaa; }")
        self._label_index.setToolTip('Index/PID')
        self._label_pid.setFont(font)
        self._label_pid.setStyleSheet("QLabel { color : #aaa; }")
        self._textedit_cmd.setStyleSheet((
            "QPlainTextEdit { "
            "background-color: #3e4041; "
            "color: #717476; "
            "border-style: none; }"
        ))
        
        # Command text edit
        self._textedit_cmd.setReadOnly(True)
        self._textedit_cmd.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self._textedit_cmd.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.MinimumExpanding
        )
        self._textedit_cmd.setFixedHeight(25)
        self._textedit_cmd.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        
        # Copy to clipboard button
        self._button_copy_cmd.setFixedSize(42, 28)
        self._button_copy_cmd.setToolTip('Copy command to clipboard')
        self._button_copy_cmd.setStyleSheet((
            "QPushButton { border-color: #666; }"
            "QPushButton:pressed { border-color: #999; }"
        ))

        self._button_copy_cmd.clicked.connect(self.copy_cmd_to_clipboard)

        # Setup layout
        glo = QtWidgets.QGridLayout()
        glo.setSpacing(1)
        glo.setContentsMargins(0, 0, 0, 0)
        glo.setColumnStretch(1, 1)
        glo.setAlignment(QtCore.Qt.AlignVCenter)
        glo.addWidget(self._label_description, 0, 0, 2, 1)
        glo.addWidget(self._label_pid, 0, 1, QtCore.Qt.AlignRight)
        glo.addWidget(self._label_update_time, 1, 1, QtCore.Qt.AlignRight)
        
        line = QtWidgets.QFrame()
        line.setGeometry(QtCore.QRect(320, 150, 118, 1))
        line.setFrameShape(QtWidgets.QFrame.HLine)
        line.setStyleSheet("QFrame { color: #555 }")
        
        glo.addWidget(line, 2, 0, 1, 3)
        
        hlo = QtWidgets.QHBoxLayout()
        hlo.setAlignment(QtCore.Qt.AlignHCenter)
        hlo.setContentsMargins(0, 0, 0, 0)
        hlo.setSpacing(0)
        hlo.addWidget(self._textedit_cmd, 0)
        hlo.addWidget(self._button_copy_cmd, 1)
        
        glo.addLayout(hlo, 3, 0, 3, 3)
        self.setLayout(glo)
    
    def update(self):
        current_item = self.view.current_item()
        
        if current_item is None:
            self._label_update_time.clear()
        else:
            time = custom_format(
                os.path.getmtime(current_item.log_path()),
                '%A %d-%m-%Y, %H:%M:%S'
            )
            self._label_update_time.setText('Last update: ' + time)
    
    def refresh(self):
        current_item = self.view.current_item()
        
        if current_item is None:
            self._label_description.clear()
            self._label_update_time.clear()
            self._label_pid.clear()
            self._textedit_cmd.clear()
        else:
            self._label_description.setText('%s - %s' % (
                current_item.name(),
                current_item.label()
            ))
            time = custom_format(
                os.path.getmtime(current_item.log_path()),
                '%A %d-%m-%Y, %H:%M:%S'
            )
            self._label_update_time.setText('Last modification: ' + time)
            self._label_pid.setText('PID: ' + str(current_item.pid()))
            self._textedit_cmd.setPlainText(current_item.command())

    def copy_cmd_to_clipboard(self):
        cmd = self._textedit_cmd.toPlainText()
        app = QtWidgets.QApplication.instance()
        clip = app.clipboard()
        clip.setText(cmd)


class OutputTextEdit(QtWidgets.QPlainTextEdit):
    
    def __init__(self, view, parent=None):
        super(OutputTextEdit, self).__init__(parent)
        self.view = view
        self._log_mtime = 0
        
        self.setReadOnly(True)
    
    def _write(self, text):
        doc = self.document()
        cursor = QtGui.QTextCursor(doc)
        cursor.clearSelection()
        cursor.movePosition(QtGui.QTextCursor.End)
        cursor.insertText(text)
        self.ensureCursorVisible()
    
    def update(self):
        self.clear()
        
        current_item = self.view.current_item()
        if current_item is None:
            return
        
        log_path = current_item.log_path()
        if log_path is None:
            return
        
        with open(log_path, 'r') as log_file:
            log = log_file.read()
            self._write(log)
    
    def refresh(self):
        current_item = self.view.current_item()
        if current_item is None:
            return
        
        log_path = current_item.log_path()
        if log_path is None:
            return

        # Check if log file has been modified
        mtime = os.path.getmtime(log_path)
        
        if mtime > self._log_mtime:
            self.update()
            self._log_mtime = mtime


class SubprocessOutput(QtWidgets.QWidget):
    
    def __init__(self, parent, view):
        super(SubprocessOutput, self).__init__(parent)
        self.view = view
        
        self._header = OutputHeader(view)
        self._output = OutputTextEdit(view)
        
        self._output.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding,
            QtWidgets.QSizePolicy.Policy.Expanding
        )
        
        vlo = QtWidgets.QVBoxLayout()
        vlo.addWidget(self._header)
        vlo.addWidget(self._output)
        vlo.setContentsMargins(0, 0, 0, 0)
        vlo.setSpacing(1)
        self.setLayout(vlo)
        
        # Set up timer to periodically update
        # running process output
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.refresh)
        self.timer.setInterval(10)
    
    def update(self):
        self._header.refresh()
        self._output.update()
    
    def refresh(self):
        self._header.update()
        self._output.refresh()


class SubprocessListItem(QtWidgets.QTreeWidgetItem):
    
    def __init__(self, tree, process):
        super(SubprocessListItem, self).__init__(tree)
        self.process = None
        # self.args = ()
        # self.kwargs = {}
        self._match_str = ''
        self.set_process(process)
    
    def set_process(self, process):
        self.process = process
        self._match_str = ''
        self._update()
    
    def __lt__(self, other):
        column = self.treeWidget().sortColumn()
        if column == 3:
            mtime = os.path.getmtime(self.log_path())
            other_mtime = os.path.getmtime(other.log_path())
            return mtime < other_mtime
        elif column == 4:
            return self.last_run_time() < other.last_run_time()
        else:
            key1 = self.text(column)
            key2 = other.text(column)
            return natural_sort_key(key1) < natural_sort_key(key2)
    
    def _update(self):
        self.setText(0, str(self.id()))
        self.setText(1, self.name())
        self.setText(2, self.label())
        mtime = os.path.getmtime(self.log_path())
        self.setText(3, timeago_format(mtime))
        self.setText(4, timeago_format(self.last_run_time()))
        self.setText(5, str(self.pid()))
        
        if self.is_running():
            color = QtGui.QColor(185, 194, 200)
        else:
            color = QtGui.QColor(110, 110, 110)
        
        for i in range(self.treeWidget().columnCount()):
            self.setForeground(i, QtGui.QBrush(color))

    def id(self):
        return self.process['id']
    
    def name(self):
        return self.process['name']
    
    def version(self):
        return self.process['version']
    
    def label(self):
        return self.process['label']
    
    def pid(self):
        return self.process['pid']
    
    def is_running(self):
        return self.process['is_running']
    
    def log_path(self):
        return self.process['log_path']
    
    def command(self):
        return self.process['command']

    def last_run_time(self):
        return self.process['last_run_time']

    def matches(self, filter):
        return filter in self._match_str


class SubprocessList(QtWidgets.QTreeWidget):
    
    def __init__(self, parent, view, session):
        super(SubprocessList, self).__init__(parent)
        self.view = view
        self.session = session
        
        columns = (
            'ID', 'Application',
            'Description', 'Last update',# 'Version',
            'Started', 'PID',
        )
        self.setHeaderLabels(columns)
        
        self.itemSelectionChanged.connect(self.view.update_output)

        # Periodically update runner infos
        self.timer = QtCore.QTimer(self)
        self.timer.timeout.connect(self.update)
        self.timer.start(1000)
        
        self.setSortingEnabled(True)
        self.sortByColumn(4, QtCore.Qt.DescendingOrder)
        
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_popup_menu_request)

        self._popup_menu = QtWidgets.QMenu(self)

        self._filter = None
        self._rid_to_item = {}
        
        self.header().resizeSection(0, 60)
        self.header().resizeSection(1, 140)
        self.header().resizeSection(3, 100)
        self.header().setSectionResizeMode(2, QtWidgets.QHeaderView.Stretch)
        self.header().setStretchLastSection(False)
    
    def set_show_running_only(self, b):
        self._show_running_only = b
        self.refresh()
    
    def refresh(self):
        #TODO: intelligent refresh: remove deleted runners, add created runners
        self.clear()
        self._rid_to_item.clear()
        for sp in self.session.cmds.SubprocessManager.list_runner_infos():
            item = SubprocessListItem(self, sp)
            self._rid_to_item[item.id()] = item
            # TODO: manage item filtering
            if self.view.show_running_only():
                item.setHidden(not item.is_running())

    def update(self):
        for sp in self.session.cmds.SubprocessManager.list_runner_infos():
            rid = sp['id']
            item = self._rid_to_item.get(rid, None)
            
            if item is None:
                # Create item for new runner instance
                item = SubprocessListItem(self, sp)
                self._rid_to_item[rid] = item
                # TODO: manage item filtering
                if self.view.show_running_only():
                    item.setHidden(not item.is_running())
            else:
                # Update info of existing runner instance
                item.set_process(sp)

    def update_runner(self, rid):
        item = self._rid_to_item[rid]
        if item is not None:
            process = self.session.cmds.SubprocessManager.get_runner_info(
                rid
            )
            item.set_process(process)
    
    def clear_completed_runners(self):
        # self.refresh()
        for i in range(self.topLevelItemCount()):
            item = self.topLevelItem(i)
            
            if not item.is_running():
                self.session.cmds.SubprocessManager.delete_runner_instance(
                    item.id()
                )
        
        self.refresh()
    
    def _on_popup_menu_request(self, pos):
        item = self.itemAt(pos)
        
        if item is None:
            m = self._popup_menu
            m.clear()
            m.addAction(
                QtGui.QIcon(resources.get_icon(('icons.libreflow', 'refresh'))),
                'Refresh',
                self.view.refresh
            )
            m.addAction(
                QtGui.QIcon(resources.get_icon(('icons.libreflow', 'clean'))),
                'Clean completed',
                self.view.clear_completed_runners
            )
        else:
            # item = item.job_item()
            m = self._popup_menu
            m.clear()
            if not item.is_running():
                m.addAction(
                    QtGui.QIcon(resources.get_icon(('icons.libreflow', 'run'))),
                    'Relaunch',
                    lambda item=item: self._launch(item)
                )
                m.addAction(
                    QtGui.QIcon(resources.get_icon(('icons.libreflow', 'delete'))),
                    'Delete',
                    lambda item=item: self._delete(item)
                )
            else:
                m.addAction(
                    QtGui.QIcon(resources.get_icon(('icons.libreflow', 'stop'))),
                    'Terminate',
                    lambda item=item: self._terminate(item)
                )

        self._popup_menu.popup(self.viewport().mapToGlobal(pos))
    
    def _launch(self, item):
        self.session.cmds.SubprocessManager.launch_runner_instance(
            item.id()
        )
        self.update_runner(item.id())
        self.view.update_output()
    
    def _terminate(self, item):
        self.session.cmds.SubprocessManager.terminate_runner_instance(
            item.id()
        )
        self.update_runner(item.id())

    def _kill(self, item):
        self.session.cmds.SubprocessManager.kill_runner_instance(
            item.id()
        )
        self.update_runner(item.id())
    
    def _delete(self, item):
        self.session.cmds.SubprocessManager.delete_runner_instance(
            item.id()
        )
        self.view.refresh()


class SubprocessView(DockedView):
    
    def __init__(self, session, view_id=None, hidden=False, area=None):
        super(SubprocessView, self).__init__(session, view_id, hidden=hidden, area=area)

    def _build(self, top_parent, top_layout, main_parent, header_parent, header_layout):
        self.splitter = QtWidgets.QSplitter(main_parent)
        self.splitter.setOrientation(QtCore.Qt.Vertical)
        
        self._subprocess_list = SubprocessList(self.splitter, self, self.session)
        self._subprocess_output = SubprocessOutput(self.splitter, self)
        
        lo = QtWidgets.QVBoxLayout()
        lo.setContentsMargins(0, 0, 0, 0)
        lo.setSpacing(0)
        lo.addWidget(self.splitter)

        main_parent.setLayout(lo)
        
        self.view_menu.setTitle('Options')
        self.view_menu.addAction(
            QtGui.QIcon(resources.get_icon(('icons.libreflow', 'refresh'))),
            'Refresh',
            self.refresh
        )
        self.view_menu.addAction(
            QtGui.QIcon(resources.get_icon(('icons.libreflow', 'clean'))),
            'Clean completed',
            self.clear_completed_runners
        )
        self._show_running_action = self.view_menu.addAction(
            'Show running only',
            self.on_show_running_change
        )
        self._show_running_action.setCheckable(True)
        self._show_running_action.setChecked(False)
        
        self.set_view_title('Processes')

    def refresh(self):
        self.refresh_list()
        self._subprocess_output.update()

    def refresh_list(self):
        self._subprocess_list.refresh()
    
    def update_output(self):
        current_item = self.current_item()
        
        if current_item is None:
            # Clear output if no runner is selected
            self._subprocess_output.timer.stop()
            self._subprocess_output.refresh()
        else:
            # Update stopped runner only if its output
            # isn't already displayed
            self._subprocess_output.timer.stop()
            self._subprocess_output.update()
            
            if current_item.is_running():
                self._subprocess_output.timer.start()
    
        # self._subprocess_output.update()
    
    def current_item(self):
        selection = self._subprocess_list.selectedItems()
        if not selection:
            return None
        else:
            return selection[0]
    
    def clear_completed_runners(self):
        self._subprocess_list.clear_completed_runners()
        self._subprocess_output.update()
    
    def on_show(self):
        self.refresh_list()

    def show_running_only(self):
        return self._show_running_action.isChecked()

    def on_show_running_change(self):
        self._subprocess_list.set_show_running_only(
            self._show_running_action.isChecked()
        )

    def receive_event(self, event_type, data):
        # TODO: Manage events sent from actor
        # (e.g. when a runner is instanciated ?)
        pass



from kabaret import flow

from libreflow import baseflow
from .lib import AssetLib
from .film import Seasons, Shot


class Project(flow.Object, flow.InjectionProvider):

    admin = flow.Child(baseflow.Admin)
    asset_library = flow.Child(AssetLib)
    seasons = flow.Child(Seasons)

    _RUNNERS_FACTORY = None

    @classmethod
    def _injection_provider(cls, slot_name, default_type):
        if slot_name == "libreflow.baseflow.film.Shot":
            return Shot



import os
import sys
import time
from pprint import pprint

from libreflow.session import BaseCLISession


def main(args):
    runner = args[0]
    oid = args[1]
    function_name = args[2]
    session = BaseCLISession(
        session_name="%s(pid=%i)" % (
            runner, os.getpid()
        ),
        debug=True,
    )
    
    session.cmds.Cluster.connect_from_env()
    session.cmds.Flow.call(oid, function_name, args={}, kwargs={})
    session.close()


if __name__ == "__main__":
    main(sys.argv[1:])
